from Products.Five import BrowserView
from zope.component.hooks import getSite


class ErrorView(BrowserView):
    def __init__(self, context, request):
        # since this is a view adapting an exception and a request (instead
        # of a regular content object and a request), the exception object
        # was passed as the context;  therefore we need to construct a
        # proper context in order to render the template in a sane manner;
        # normally we could use `getUtility(ISiteRoot)` for that, but then
        # `REQUEST` is missing, we'll employ the getSite hook ...
        self.exception = context
        self.context = getSite()
        self.request = request

    def errorInfo(self):
        """provide error type and value"""
        type_ = str(self.exception.__class__).replace("<class '", "").replace("'>", "")
        return {"type": type_, "value": self.exception.args}



from unittest import TestCase

from collective.solr.browser.facets import SearchBox, SearchFacetsView
from collective.solr.browser.interfaces import IThemeSpecific
from collective.solr.dispatcher import solrSearchResults
from collective.solr.exceptions import SolrConnectionException
from collective.solr.testing import (
    LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING,
    activateAndReindex,
)
from collective.solr.utils import activate
from plone import api
from plone.app.testing import TEST_USER_ID, setRoles
from zope.component import getMultiAdapter
from zope.interface import alsoProvides


class SolrFacettingTests(TestCase):
    layer = LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        activateAndReindex(self.portal)
        response = self.request.RESPONSE
        self.write = response.write
        response.write = lambda x: x  # temporarily ignore output
        self.maintenance = self.portal.unrestrictedTraverse("solr-maintenance")

    def tearDown(self):
        self.request.RESPONSE.write = self.write
        activate(active=False)

    def _create_uf(self):
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        if "Members" not in self.portal:
            mf = api.content.create(self.portal, "Folder", id="Members")
        else:
            mf = self.portal["Members"]
        if TEST_USER_ID not in mf:
            uf = api.content.create(mf, "Folder", id=TEST_USER_ID)
        else:
            uf = mf[TEST_USER_ID]
            api.content.transition(obj=mf, transition="hide")
            api.content.transition(obj=uf, transition="hide")
        self.maintenance.reindex()
        setRoles(self.portal, TEST_USER_ID, [])

    def testFacettedSearchWithKeywordArguments(self):
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Event", id="event1", title="Welcome")
        self.maintenance.reindex()
        results = solrSearchResults(
            SearchableText="Welcome", facet="true", facet_field="portal_type"
        )
        self.assertEqual(
            sorted([r.path_string for r in results]),
            ["/plone/event1", "/plone/front-page"],
        )
        types = results.facet_counts["facet_fields"]["portal_type"]
        self.assertEqual(types["Document"], 1)
        self.assertEqual(types["Event"], 1)

    def testFacettedSearchWithRequestArguments(self):
        self._create_uf()
        self.request = self.layer["request"]
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "review_state"
        results = solrSearchResults(self.request)
        self.assertEqual(
            sorted([r.path_string for r in results]),
            ["/plone/news", "/plone/news/aggregator"],
        )
        states = results.facet_counts["facet_fields"]["review_state"]
        self.assertEqual(states, dict(private=0, published=2))

    def testMultiFacettedSearch(self):
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Event", id="event1", title="Welcome")
        self.maintenance.reindex()
        results = solrSearchResults(
            SearchableText="Welcome",
            facet="true",
            facet_field=["portal_type", "review_state"],
        )
        self.assertEqual(
            sorted([r.path_string for r in results]),
            ["/plone/event1", "/plone/front-page"],
        )
        facets = results.facet_counts["facet_fields"]
        self.assertEqual(facets["portal_type"]["Event"], 1)
        self.assertEqual(facets["review_state"]["published"], 1)

    def testFacettedSearchWithFilterQuery(self):
        self._create_uf()
        self.request = self.layer["request"]
        self.request.form["SearchableText"] = "News"
        self.request.form["fq"] = "portal_type:Collection"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "review_state"
        results = solrSearchResults(self.request)
        self.assertEqual([r.path_string for r in results], ["/plone/news/aggregator"])
        states = results.facet_counts["facet_fields"]["review_state"]
        self.assertEqual(states, dict(private=0, published=1))

    def testFacettedSearchWithDependencies(self):
        # facets depending on others should not show up initially
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = ["portal_type", "review_state:portal_type"]
        view = SearchFacetsView(self.portal, self.request)
        view.kw = dict(results=solrSearchResults(self.request))
        facets = [facet["title"] for facet in view.facets()]
        self.assertEqual(facets, ["portal_type"])
        # now again with the required facet selected
        self.request.form["fq"] = "portal_type:Collection"
        view.kw = dict(results=solrSearchResults(self.request))
        facets = [facet["title"] for facet in view.facets()]
        self.assertEqual(facets, ["portal_type", "review_state"])

    def testFacettedSearchWithUnicodeFilterQuery(self):
        self.portal.news.portal_type = "Føø"
        self.maintenance.reindex()
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "portal_type"
        view = SearchFacetsView(self.portal, self.request)
        view.kw = dict(results=solrSearchResults(self.request))
        facets = view.facets()
        self.assertEqual(
            sorted([c["name"] for c in facets[0]["counts"]]), ["Collection", u"Føø"]
        )

    def checkOrder(self, html, *order):
        for item in order:
            position = html.find(item)
            self.assertTrue(
                position >= 0, 'menu item "%s" missing or out of order' % item
            )
            html = html[position:]

    def testFacetsInformationView(self):
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Event", id="event1", title="Welcome")
        self.maintenance.reindex()
        self.request.form["SearchableText"] = "Welcome"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "portal_type"
        alsoProvides(self.request, IThemeSpecific)
        view = getMultiAdapter((self.portal, self.request), name="search-facets")
        if hasattr(view, "__of__"):
            view = view.__of__(self.portal)
        results = solrSearchResults(self.request)
        output = view(results=results)
        self.checkOrder(
            output, "portal-searchfacets", "Content type", "Document", "1", "Event", "1"
        )

    def testFacetFieldsInSearchBox(self):
        self.request = self.portal.REQUEST
        viewlet = SearchBox(self.portal, self.request, None, None)
        if hasattr(viewlet, "__of__"):
            viewlet = viewlet.__of__(self.portal)
        viewlet.update()
        output = viewlet.render()
        self.checkOrder(
            output,
            "<input",
            'name="facet" value="true"',
            "<input",
            'value="portal_type"',
            "<input",
            'value="review_state"',
            "</form>",
        )
        # try overriding the desired facets in the self.request
        self.request.form["facet.field"] = ["foo"]
        output = viewlet.render()
        self.checkOrder(
            output,
            "<input",
            'name="facet" value="true"',
            "<input",
            'value="foo"',
            "</form>",
        )
        self.assertFalse("portal_type" in output)

    def testUnknownFacetField(self):
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "foo"
        alsoProvides(self.request, IThemeSpecific)
        self.assertRaises(SolrConnectionException, solrSearchResults, self.request)

    def testNoFacetFields(self):
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = []
        alsoProvides(self.request, IThemeSpecific)
        view = getMultiAdapter((self.portal, self.request), name="search-facets")
        if hasattr(view, "__of__"):
            view = view.__of__(self.portal)
        output = view(results=solrSearchResults(self.request))
        self.assertFalse("portal-searchfacets" in output, output)

    def testEmptyFacetValue(self):
        # let's artificially create an empty value;  while this is a
        # somewhat unrealistic scenario, empty values may very well occur
        # for additional custom indexes...
        self.portal.news.portal_type = ""
        self.maintenance.reindex()
        # after updating the solr index the view can be checked...
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = "portal_type"
        alsoProvides(self.request, IThemeSpecific)
        view = getMultiAdapter((self.portal, self.request), name="search-facets")
        if hasattr(view, "__of__"):
            view = view.__of__(self.portal)
        results = solrSearchResults(self.request)
        output = view(results=results)
        # the empty facet value should be displayed resulting in
        # only one list item (`<dd>`)
        self.assertEqual(len(output.split("<dd>")), 2)
        # let's also make sure there are no empty filter queries
        self.assertFalse("fq=portal_type%3A&amp;" in output)

    def testFacetOrder(self):
        self.request.form["SearchableText"] = "News"
        self.request.form["facet"] = "true"
        self.request.form["facet_field"] = ["portal_type", "review_state"]
        alsoProvides(self.request, IThemeSpecific)
        view = getMultiAdapter((self.portal, self.request), name="search-facets")
        if hasattr(view, "__of__"):
            view = view.__of__(self.portal)
        results = solrSearchResults(self.request)
        output = view(results=results)
        # the displayed facets should match the given order...
        self.checkOrder(output, "portal-searchfacets", "Content type", "Review state")
        # let's also try the other way round...
        self.request.form["facet_field"] = ["review_state", "portal_type"]
        results = solrSearchResults(self.request)
        output = view(results=results)
        self.checkOrder(output, "portal-searchfacets", "Review state", "Content type")



import unittest

from collective.solr.dispatcher import solrSearchResults
from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.testing import (
    LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING,
    activateAndReindex,
)
from collective.solr.utils import activate, getConfig
from plone.app.testing import TEST_USER_ID, setRoles
from plone.dexterity.content import Item
from plone.dexterity.fti import DexterityFTI
from Products.CMFCore.utils import getToolByName
from transaction import commit
from zope import schema
from zope.component import getUtility
from zope.interface import Interface, implementer


class ILocation(Interface):

    geolocation = schema.TextLine(
        title=u"Geo Location", description=u"", required=False
    )


@implementer(ILocation)
class Location(Item):
    """ """


class SolrSpatialSearchTests(unittest.TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self._setUpGeolocationCatalogIndex()
        self._setUpLocationType()
        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.config = getConfig()
        self.config.required = []
        self.maintenance = self.portal.unrestrictedTraverse("solr-maintenance")
        commit()
        self.response = self.request.RESPONSE
        self.write = self.response.write
        self.response.write = lambda x: x  # temporarily ignore output
        activateAndReindex(self.portal)

    def tearDown(self):
        activate(active=False)
        self.response.write = self.write

    def _setUpLocationType(self):
        types_tool = getToolByName(self.portal, "portal_types")
        fti = DexterityFTI("Location")
        fti.klass = "collective.solr.tests.test_spatial.Location"
        types_tool._setObject("Location", fti)

    def _setUpGeolocationCatalogIndex(self):
        catalog = getToolByName(self.portal, "portal_catalog")
        catalog.addIndex("geolocation", "FieldIndex")

    def _makeSearchableTextOptionalInSolrConfiguration(self):
        config = getUtility(ISolrConnectionConfig)
        config.required = ()

    def testGeoSpatialTypeIsIndexed(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(portal_type="Location")
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])

    def testGeoSpatialFieldCanBeDeleted(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            sfield="geolocation",
            fq="{!geofilt}",
            pt="50.2,-7.1",
            d=0,
        )
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])
        self.portal.location1.geolocation = None
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            sfield="geolocation",
            fq="{!geofilt}",
            pt="50.2,-7.1",
            d=0,
        )
        self.assertEqual(list(results), [])
        results = solrSearchResults(portal_type="Location")
        from Missing import Missing

        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])
        self.assertTrue(isinstance(results[0].geolocation, Missing))

    def testGeoSpatialFieldCanBeUpdated(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            sfield="geolocation",
            fq="{!geofilt}",
            pt="50.2,-7.1",
            d=0,
        )
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])
        self.portal.location1.geolocation = "10.1, -8.4"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            sfield="geolocation",
            fq="{!geofilt}",
            pt="10.1, -8.4",
            d=0,
        )
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])

    def testGeoSpatialSearchWithExactLocation(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            sfield="geolocation",
            fq="{!geofilt}",
            pt="50.2,-7.1",
            d=0,
        )
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])

    def testGeoSpatialSearchWithNearbyLocation(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            portal_type="Location",
            fq="{!geofilt}",
            sfield="geolocation",
            pt="50.3,-7.4",
            d=100,
        )
        self.assertEqual(sorted([r.path_string for r in results]), ["/plone/location1"])

    def testGeoSpatialSearchWithLocationOutsideTheSearchBox(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50.2,-7.1"
        self.portal.location1.reindexObject()
        self.maintenance.reindex()
        results = solrSearchResults(
            fq="{!geofilt}", sfield="geolocation", pt="80,123", d=0.2
        )
        self.assertEqual(sorted([r.path_string for r in results]), [])

    # sort='geodist() desc' allows us to sort geolocations by distance.
    # it seems that right now collective.solr filters that param entirely.
    @unittest.skip("Sorting is not working yet.")
    def testGeoSpatialSearchSortByDistance(self):
        self.portal.invokeFactory("Location", id="location1", title="Loc 1")
        self.portal.location1.geolocation = "50,1"
        self.portal.location1.reindexObject()
        self.portal.invokeFactory("Location", id="location2", title="Loc 2")
        self.portal.location2.geolocation = "60,1"
        self.portal.location2.reindexObject()
        self.maintenance.reindex()

        # Query: 52,1 | 50,1 (locaction1) < 60,1 (location2)
        results = solrSearchResults(
            fq="{!bbox}", sfield="geolocation", pt="52,1", d=100, sort="geodist() desc"
        )
        self.assertEqual(
            sorted([r.path_string for r in results]),
            ["/plone/location2", "/plone/location1"],
        )

        # Query: 58,1 | 60,1 (locaction2) < 50,1 (location1)
        results = solrSearchResults(
            fq="{!bbox}", sfield="geolocation", pt="58,1", d=100, sort="geodist() asc"
        )
        self.assertEqual(
            sorted([r.path_string for r in results]),
            ["/plone/location1", "/plone/location2"],
        )


# def test_suite():
#     if pingSolr():
#         return defaultTestLoader.loadTestsFromName(__name__)
#     else:
#         return TestSuite()



from __future__ import print_function

from os.path import dirname, join
from re import search
from socket import error
from sys import stderr
from threading import Thread

import six
from collective.solr import tests
from collective.solr.local import getLocal, setLocal
from six.moves.BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from six.moves.http_client import HTTPConnection
from zope.component.hooks import getSite, setSite

try:
    from Zope2.App import zcml
except ImportError:
    from Products.Five import zcml


def loadZCMLString(string):
    # Unset current site for Zope 2.13
    saved = getSite()
    setSite(None)
    try:
        zcml.load_string(string)
    finally:
        setSite(saved)


def getData(filename):
    """return a file object from the test data folder"""
    filename = join(dirname(tests.__file__), "data", filename)
    return open(filename, "rb").read()


def fakehttp(solrconn, *fakedata):
    """helper function to set up a fake http request on a SolrConnection"""

    class FakeOutput(list):

        """helper class to organize output from fake connections"""

        conn = solrconn

        def log(self, item):
            self.current.append(item)

        def get(self, skip=0):
            self[:] = self[skip:]
            return b"".join(self.pop(0)).replace(b"\r", b"")

        def new(self):
            self.current = []
            self.append(self.current)

        def __len__(self):
            self.conn.flush()  # send out all pending xml
            return super(FakeOutput, self).__len__()

        def __str__(self):
            self.conn.flush()  # send out all pending xml
            if self:
                return "".join([chunk.decode("utf-8") for chunk in self[0]]).replace(
                    "\r", ""
                )
            else:
                return ""

    output = FakeOutput()

    class FakeSocket(six.BytesIO):

        """helper class to fake socket communication"""

        def sendall(self, str):
            output.log(str)

        if six.PY2:

            def makefile(self, mode, name):
                return self

        else:

            def makefile(self, mode):
                return self

        def read(self, amt=None):
            if self.closed:
                return b""
            return six.BytesIO.read(self, amt)

        def readline(self, length=None):
            if self.closed:
                return b""
            return six.BytesIO.readline(self, length)

    class FakeHTTPConnection(HTTPConnection):

        """helper class to fake a http connection object from httplib.py"""

        def __init__(self, host, *fakedata):
            HTTPConnection.__init__(self, host)
            self.fakedata = list(fakedata)

        def putrequest(self, *args, **kw):
            self.url = args[1]
            response = self.fakedata.pop(0)  # get first response
            self.sock = FakeSocket(response)  # and set up a fake socket
            output.new()  # as well as an output buffer
            HTTPConnection.putrequest(self, *args, **kw)

        def setTimeout(self, timeout):
            pass

    solrconn.conn = FakeHTTPConnection(solrconn.conn.host, *fakedata)
    return output


def fakemore(solrconn, *fakedata):
    """helper function to add more fake http requests to a SolrConnection"""
    assert hasattr(solrconn.conn, "fakedata")  # `isinstance()` doesn't work?
    solrconn.conn.fakedata.extend(fakedata)


def fakeServer(actions, port=55555):
    """helper to set up and activate a fake http server used for testing
    purposes; <actions> must be a list of handler functions, which will
    receive the base handler as their only argument and are used to
    process the incoming requests in turn; returns a thread that should
    be 'joined' when done"""

    class Handler(BaseHTTPRequestHandler):
        def do_POST(self):
            action = actions.pop(0)  # get next action
            action(self)  # and process it...

        def do_GET(self):
            action = actions.pop(0)  # get next action
            action(self)  # and process it...

        def log_request(*args, **kw):
            pass

    def runner():
        while actions:
            server.handle_request()

    server = HTTPServer(("", port), Handler)
    thread = Thread(target=runner)
    thread.start()
    return thread


def pingSolr():
    """test if the solr server is available"""
    status = getLocal("solrStatus")
    if status is not None:
        return status
    conn = HTTPConnection("localhost", 8983)
    try:
        conn.request("GET", "/solr/plone/admin/ping")
        response = conn.getresponse()
        status = response.status == 200
        msg = "INFO: solr return status '%s'" % response.status
    except error as e:
        status = False
        msg = 'WARNING: solr tests could not be run: "%s".' % e
    if not status:
        print(file=stderr)
        print("*" * len(msg), file=stderr)
        print(msg, file=stderr)
        print("*" * len(msg), file=stderr)
        print(file=stderr)
    setLocal("solrStatus", status)
    return status


def numFound(result):
    if isinstance(result, six.binary_type):
        result = result.decode("utf-8")
    match = search(r'numFound="(\d+)"', result)
    if match is not None:
        match = int(match.group(1))
    return match



from logging import getLogger
from time import strftime, time

from BTrees.IIBTree import IITreeSet
from collective.solr.flare import PloneFlare
from collective.solr.indexer import DefaultAdder, SolrIndexProcessor, boost_values
from collective.solr.interfaces import (
    ICheckIndexable,
    ISolrAddHandler,
    ISolrConnectionManager,
    ISolrMaintenanceView,
    MAX_RESULTS_SUPPORTED_BY_SOLR,
)
from collective.solr.parser import SolrResponse, parse_date_as_datetime, unmarshallers
from collective.solr.utils import findObjects, prepareData
from plone.uuid.interfaces import IUUID, IUUIDAware
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from zope.component import queryAdapter, queryUtility
from zope.interface import implementer

logger = getLogger("collective.solr.maintenance")


try:
    from time import process_time
except ImportError:
    # Python < 3.8
    from time import clock as process_time


def timer(func=time):
    """set up a generator returning the elapsed time since the last call"""

    def gen(last=func()):
        while True:
            elapsed = func() - last
            last = func()
            yield "%.3fs" % elapsed

    return gen()


def checkpointIterator(function, interval=100):
    """the iterator will call the given function for every nth invocation"""
    counter = 0
    while True:
        counter += 1
        if counter % interval == 0:
            function()
        yield None


def notimeout(func):
    """decorator to prevent long-running solr tasks from timing out"""

    def wrapper(*args, **kw):
        """wrapper with random docstring so ttw access still works"""
        manager = queryUtility(ISolrConnectionManager)
        manager.setTimeout(None, lock=True)
        try:
            return func(*args, **kw)
        finally:
            manager.setTimeout(None, lock=False)

    return wrapper


@implementer(ISolrMaintenanceView)
class SolrMaintenanceView(BrowserView):
    """helper view for indexing all portal content in Solr"""

    def mklog(self, use_std_log=False):
        """helper to prepend a time stamp to the output"""
        write = self.request.RESPONSE.write

        def log(msg, timestamp=True):
            if timestamp:
                msg = strftime("%Y/%m/%d-%H:%M:%S ") + msg
            write(msg.encode("utf-8"))
            if use_std_log:
                logger.info(msg)

        return log

    def optimize(self):
        """optimize solr indexes"""
        manager = queryUtility(ISolrConnectionManager)
        conn = manager.getConnection()
        conn.setTimeout(None)
        conn.commit(optimize=True)
        return "solr indexes optimized."

    def clear(self):
        """clear all data from solr, i.e. delete all indexed objects"""
        manager = queryUtility(ISolrConnectionManager)
        uniqueKey = manager.getSchema().uniqueKey
        conn = manager.getConnection()
        conn.setTimeout(None)
        conn.deleteByQuery("%s:[* TO *]" % uniqueKey)
        conn.commit()
        return "solr index cleared."

    def reindex(
        self,
        batch=1000,
        skip=0,
        limit=0,
        ignore_portal_types=None,
        only_portal_types=None,
        idxs=[],
        ignore_exceptions=False,
    ):
        """find all contentish objects (meaning all objects derived from one
        of the catalog mixin classes) and (re)indexes them"""

        if ignore_portal_types and only_portal_types:
            raise ValueError(
                "It is not possible to combine "
                "ignore_portal_types with only_portal_types"
            )

        atomic = idxs != []
        manager = queryUtility(ISolrConnectionManager)
        proc = SolrIndexProcessor(manager)
        conn = manager.getConnection()
        zodb_conn = self.context._p_jar
        log = self.mklog()
        log("reindexing solr catalog...\n")
        if skip:
            log("skipping indexing of %d object(s)...\n" % skip)
        if limit:
            log("limiting indexing to %d object(s)...\n" % limit)
        real = timer()  # real time
        lap = timer()  # real lap time (for intermediate commits)
        cpu = timer(process_time)  # cpu time
        processed = 0
        schema = manager.getSchema()
        key = schema.uniqueKey
        updates = {}  # list to hold data to be updated

        def flush():
            return conn.commit(soft=True)

        flush = notimeout(flush)

        def checkPoint():
            for my_boost_values, data in updates.values():
                adder = data.pop("_solr_adder")
                try:
                    adder(conn, boost_values=my_boost_values, **data)
                except Exception as e:
                    logger.warning("Error %s @ %s", e, data["path_string"])
                    if not ignore_exceptions:
                        raise
            updates.clear()
            msg = (
                "intermediate commit (%d items processed, "
                "last batch in %s)...\n" % (processed, next(lap))
            )
            log(msg)
            logger.info(msg)
            flush()
            zodb_conn.cacheGC()

        cpi = checkpointIterator(checkPoint, batch)
        count = 0

        if atomic:
            log("indexing only {0} \n".format(idxs))

        for path, obj in findObjects(self.context):
            if ICheckIndexable(obj)():
                count += 1
                if count <= skip:
                    continue

                if ignore_portal_types:
                    if obj.portal_type in ignore_portal_types:
                        continue

                if only_portal_types:
                    if obj.portal_type not in only_portal_types:
                        continue

                attributes = None
                if atomic:
                    attributes = idxs

                # For atomic updates to work the uniqueKey must be present
                # in *every* update operation.
                if attributes and key not in attributes:
                    attributes.append(key)
                data, missing = proc.getData(obj, attributes=attributes)
                prepareData(data)

                if not missing or atomic:
                    value = data.get(key, None)
                    if value is not None:
                        log("indexing %r\n" % obj)

                        pt = data.get("portal_type", "default")
                        adder = queryAdapter(obj, ISolrAddHandler, name=pt)
                        if adder is None:
                            adder = DefaultAdder(obj)
                        data["_solr_adder"] = adder
                        updates[value] = (boost_values(obj, data), data)
                        processed += 1
                        next(cpi)
                else:
                    log("missing data, skipping indexing of %r.\n" % obj)
                if limit and count >= (skip + limit):
                    break

        checkPoint()
        conn.commit()
        log("solr index rebuilt.\n")
        msg = "processed %d items in %s (%s cpu time)."
        msg = msg % (processed, next(real), next(cpu))
        log(msg)
        logger.info(msg)

    def sync(self, batch=1000, preImportDeleteQuery="*:*"):
        """Sync the Solr index with the portal catalog. Records contained
        in the catalog but not in Solr will be indexed and records not
        contained in the catalog will be removed.
        """
        manager = queryUtility(ISolrConnectionManager)
        proc = SolrIndexProcessor(manager)
        conn = manager.getConnection()
        key = queryUtility(ISolrConnectionManager).getSchema().uniqueKey
        zodb_conn = self.context._p_jar
        catalog = getToolByName(self.context, "portal_catalog")
        getIndex = catalog._catalog.getIndex
        modified_index = getIndex("modified")
        uid_index = getIndex(key)
        log = self.mklog()
        real = timer()  # real time
        lap = timer()  # real lap time (for intermediate commits)
        cpu = timer(process_time)  # cpu time
        # get Solr status
        response = conn.search(
            q=preImportDeleteQuery,
            rows=MAX_RESULTS_SUPPORTED_BY_SOLR,
            fl="%s modified" % key,
        )
        # avoid creating DateTime instances
        simple_unmarshallers = unmarshallers.copy()
        simple_unmarshallers["date"] = parse_date_as_datetime
        flares = SolrResponse(response, simple_unmarshallers)
        response.close()
        solr_results = {}
        solr_uids = set()

        def _utc_convert(value):
            t_tup = value.utctimetuple()
            return (
                ((t_tup[0] * 12 + t_tup[1]) * 31 + t_tup[2]) * 24 + t_tup[3]
            ) * 60 + t_tup[4]

        for flare in flares:
            uid = flare[key]
            solr_uids.add(uid)
            solr_results[uid] = _utc_convert(flare["modified"])
        # get catalog status
        cat_results = {}
        cat_uids = set()
        for uid, rid in uid_index._index.items():
            cat_uids.add(uid)
            cat_results[uid] = rid
        # differences
        index = cat_uids.difference(solr_uids)
        solr_uids.difference_update(cat_uids)
        unindex = solr_uids
        processed = 0
        flush = notimeout(lambda: conn.flush())

        def checkPoint():
            msg = (
                "intermediate commit (%d items processed, "
                "last batch in %s)...\n" % (processed, next(lap))
            )
            log(msg)
            logger.info(msg)
            flush()
            zodb_conn.cacheGC()

        cpi = checkpointIterator(checkPoint, batch)
        # Look up objects
        uid_rid_get = cat_results.get
        rid_path_get = catalog._catalog.paths.get
        catalog_traverse = catalog.unrestrictedTraverse

        def lookup(
            uid,
            rid=None,
            uid_rid_get=uid_rid_get,
            rid_path_get=rid_path_get,
            catalog_traverse=catalog_traverse,
        ):
            if rid is None:
                rid = uid_rid_get(uid)
            if not rid:
                return None
            if not isinstance(rid, int):
                rid = tuple(rid)[0]
            path = rid_path_get(rid)
            if not path:
                return None
            try:
                obj = catalog_traverse(path)
            except AttributeError:
                return None
            return obj

        log('processing %d "unindex" operations next...\n' % len(unindex))
        op = notimeout(lambda uid: conn.delete(id=uid))
        for uid in unindex:
            obj = lookup(uid)
            if obj is None:
                op(uid)
                processed += 1
                next(cpi)
            else:
                log("not unindexing existing object %r.\n" % uid)
        log('processing %d "index" operations next...\n' % len(index))
        op = notimeout(lambda obj: proc.index(obj))
        for uid in index:
            obj = lookup(uid)
            if ICheckIndexable(obj)():
                op(obj)
                processed += 1
                next(cpi)
            else:
                log("not indexing unindexable object %r.\n" % uid)
            if obj is not None:
                obj._p_deactivate()
        log('processing "reindex" operations next...\n')
        op = notimeout(lambda obj: proc.reindex(obj))
        cat_mod_get = modified_index._unindex.get
        solr_mod_get = solr_results.get
        done = unindex.union(index)
        for uid, rid in cat_results.items():
            if uid in done:
                continue
            if isinstance(rid, IITreeSet):
                rid = list(rid.keys())[0]
            if cat_mod_get(rid) != solr_mod_get(uid):
                obj = lookup(uid, rid=rid)
                if ICheckIndexable(obj)():
                    op(obj)
                    processed += 1
                    next(cpi)
                else:
                    log("not reindexing unindexable object %r.\n" % uid)
                if obj is not None:
                    obj._p_deactivate()
        conn.commit()
        log("solr index synced.\n")
        msg = "processed %d object(s) in %s (%s cpu time)."
        msg = msg % (processed, next(real), next(cpu))
        log(msg)
        logger.info(msg)

    def cleanup(self, batch=1000):
        """remove entries from solr that don't have a corresponding Zope
        object or have a different UID than the real object"""
        manager = queryUtility(ISolrConnectionManager)
        proc = SolrIndexProcessor(manager)
        conn = manager.getConnection()
        log = self.mklog(use_std_log=True)
        log("cleaning up solr index...\n")
        key = manager.getSchema().uniqueKey

        start = 0
        resp = SolrResponse(conn.search(q="*:*", rows=batch, start=start))
        res = resp.results()
        log("%s items in solr catalog\n" % resp.response.numFound)
        deleted = 0
        reindexed = 0
        while len(res) > 0:
            for flare in res:
                try:
                    ob = PloneFlare(flare).getObject()
                except Exception as err:
                    log(
                        "Error getting object, removing: %s (%s)\n"
                        % (flare["path_string"], err)
                    )
                    conn.delete(flare[key])
                    deleted += 1
                    continue
                if ob is None:
                    log("Object not found, removing: %s\n" % (flare["path_string"]))
                    conn.delete(flare[key])
                    deleted += 1
                    continue
                if not IUUIDAware.providedBy(ob):
                    no_skipping_msg = (
                        "Object %s of type %s does not " + "support uuids, skipping.\n"
                    )
                    log(
                        no_skipping_msg % ("/".join(ob.getPhysicalPath()), ob.meta_type)
                    )
                    continue
                uuid = IUUID(ob)
                if uuid != flare[key]:
                    log(
                        "indexed under wrong UID, removing: %s\n" % flare["path_string"]
                    )
                    conn.delete(flare[key])
                    deleted += 1
                    realob_res = SolrResponse(
                        conn.search(q="%s:%s" % (key, uuid))
                    ).results()
                    if len(realob_res) == 0:
                        log("no sane entry for last object, reindexing\n")
                        data, missing = proc.getData(ob)
                        prepareData(data)
                        if not missing:
                            boost = boost_values(ob, data)
                            conn.add(boost_values=boost, **data)
                            reindexed += 1
                        else:
                            log("  missing data, cannot index.\n")
            log("handled batch of %d items, committing\n" % len(res))
            conn.commit()
            start += batch
            resp = SolrResponse(conn.search(q="*:*", rows=batch, start=start))
            res = resp.results()
        finished_msg = (
            "solr cleanup finished, %s item(s) removed, " + "%s item(s) reindexed\n"
        )
        msg = finished_msg % (deleted, reindexed)
        log(msg)
        logger.info(msg)



import re

from collective.solr.utils import getConfig

reLine = re.compile(r"^\s*([A-Za-zÀ-ÖØ-öø-ÿ]*)")

raw = None
raw_case_insensitive = None
cooked = None


def parseStopwords(stopwords, stopwords_case_insensitive):
    return list(
        map(
            lambda word: word.lower() if stopwords_case_insensitive else word,
            filter(
                lambda word: word,
                map(lambda line: reLine.match(line).group(1), stopwords.splitlines()),
            ),
        )
    )


def getStopWords(config):
    global raw
    global cooked
    global raw_case_insensitive
    config = config or getConfig()
    stopwords = getattr(config, "stopwords", "") or ""
    stopwords_case_insensitive = getattr(config, "stopwords_case_insensitive", False)
    if (
        cooked is None
        or raw is not stopwords
        or raw_case_insensitive != stopwords_case_insensitive
    ):
        raw = stopwords
        raw_case_insensitive = stopwords_case_insensitive
        cooked = parseStopwords(raw, stopwords_case_insensitive)
    return cooked


def isStopWord(term, config):
    stopwords_case_insensitive = getattr(config, "stopwords_case_insensitive", False)
    stopwords = getStopWords(config)
    return (term.lower() if stopwords_case_insensitive else term) in stopwords



import unittest

from collective.solr.behaviors import ISolrFields
from collective.solr.testing import (
    LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING,
    activateAndReindex,
)
from plone.app.testing import TEST_USER_ID, setRoles
from plone.dexterity.fti import DexterityFTI


class BehaviorsTestCase(unittest.TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        activateAndReindex(self.portal)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        fti = DexterityFTI("My Dexterity Container")
        self.portal.portal_types._setObject("My Dexterity Container", fti)
        fti.klass = "plone.dexterity.content.Container"
        fti.schema = "zope.interface.Interface"
        fti.behaviors = ("collective.solr.behaviors.ISolrFields",)
        self.portal.invokeFactory("My Dexterity Container", "child")
        self.child = self.portal["child"]

    def test_behavior_showinsearch(self):
        self.assertTrue(ISolrFields(self.child).showinsearch)

    def test_behavior_searchwords(self):
        self.assertEqual(ISolrFields(self.child).searchwords, [])


# EOF



class SolrException(Exception):
    """Base class where all other exceptions from collective.solr derive from"""


class SolrInactiveException(SolrException):
    """An exception indicating the solr integration is not activated"""


class FallBackException(SolrException):
    """Exception indicating the dispatcher should fall back to searching
    the portal catalog
    """


class SolrConnectionException(SolrException):
    """An exception thrown by solr connections"""

    def __init__(self, httpcode="000", reason=None, body=None):
        self.httpcode = httpcode
        self.reason = reason
        self.body = body

    def __repr__(self):
        return "HTTP code=%s, Reason=%s, body=%s" % (
            self.httpcode,
            self.reason,
            self.body,
        )

    def __str__(self):
        return "HTTP code=%s, reason=%s" % (self.httpcode, self.reason)


class RepeatableSolrConnectionException(SolrConnectionException):
    """An exception thrown by solr connections, that can be repeated once"""



import json

import six.moves.urllib.error
import six.moves.urllib.parse
import six.moves.urllib.request
from collective.solr.interfaces import ISolrConnectionManager
from Products.Five.browser import BrowserView
from zope.component import getUtility


class SuggestView(BrowserView):
    def __call__(self):
        suggestions = []
        term = self.request.get("term", "")
        if not term:
            return json.dumps(suggestions)
        manager = getUtility(ISolrConnectionManager)
        connection = manager.getConnection()

        if connection is None:
            return json.dumps(suggestions)

        params = {}
        params["q"] = term
        params["wt"] = "json"

        params = six.moves.urllib.parse.urlencode(params, doseq=True)
        response = connection.doGet(connection.solrBase + "/spell?" + params, {})
        results = json.loads(response.read())

        # Check for spellcheck
        spellcheck = results.get("spellcheck", None)
        if not spellcheck:
            return json.dumps(suggestions)

        # Check for existing spellcheck suggestions
        spellcheck_suggestions = spellcheck.get("suggestions", None)
        correctly_spelled = spellcheck_suggestions == [u"correctlySpelled", True]

        # Autocomplete
        if correctly_spelled:
            return json.dumps([x["Title"] for x in results["response"]["docs"]])

        if not spellcheck_suggestions:
            return json.dumps(suggestions)

        # Collect suggestions
        if spellcheck_suggestions[1]:
            for suggestion in spellcheck_suggestions[1]["suggestion"]:
                suggestions.append(dict(label=suggestion, value=suggestion))

        return json.dumps(suggestions)


class AutocompleteView(BrowserView):
    def __call__(self):
        term = self.request.get("term", "")
        if not term:
            return json.dumps([])
        manager = getUtility(ISolrConnectionManager)
        connection = manager.getConnection()

        if connection is None:
            return json.dumps([])

        params = {}
        params["q"] = term
        params["wt"] = "json"

        params = six.moves.urllib.parse.urlencode(params, doseq=True)
        url = connection.solrBase + "/autocomplete?" + params
        response = connection.doGet(url, {})
        results = json.loads(response.read())

        if "grouped" not in results:
            return json.dumps([])

        groups = results.get("grouped")["title_autocomplete"]["groups"]

        suggestions = [x["doclist"]["docs"][0]["title_autocomplete"] for x in groups]

        result = []
        for suggestion in suggestions:
            result.append(dict(label=suggestion, value=suggestion))
        return json.dumps(result)



from re import compile

import six

# Solr/lucene reserved characters/terms:
#   + - && || ! ( ) { } [ ] ^ " ~ * ? : \ /
#   (see http://wiki.apache.org/solr/SolrQuerySyntax)
# Four groups for tokenizer:
# 1) Whitespace (\s+)
# 2) Boolean operators, ampersand and pipe literals ([\&\|])
# 2) Any non reserved characters (normal text) ([^-(){}\[\]+!^\"~*?:\\/\s]+)
# 3) Any grouping characters ([(){}\[\]"])
# 4) Any special operators ([-+!^~*?:\\/])
query_tokenizer = compile(
    r"(?:(\s+)|"
    r"([\&\|])|"
    r"([^-(){}\[\]+!^\"~*?:\\/\&\|\s]+)|"
    r'([(){}\[\]"])|'
    r"([-+!^~*?:\\/]))"
)


class Whitespace(object):
    def __nonzero__(self):
        return False

    def __str__(self):
        return " "


class Group(list):
    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end
        self.isgroup = False  # Set on pop

    def __str__(self):
        res = [x for x in self if x]
        lenres = len(res)
        if lenres == 0:
            return ""
        elif lenres == 1:
            return six.text_type(res[0])
        # Otherwise, also print whitespace
        return "%s%s%s" % (
            self.start,
            "".join([six.text_type(x) for x in self]),
            self.end,
        )


class Quote(Group):
    def __str__(self):
        if not self.end:
            # No finishing quote, we have to add new group if there is
            # whitespace
            if [x for x in self if isinstance(x, Whitespace)]:
                self.start = "(%s" % self.start
                self.end = ")"
        return "%s%s%s" % (
            self.start,
            "".join([six.text_type(x) for x in self]),
            self.end,
        )


class Range(Group):
    def __str__(self):
        first = last = "*"
        if len(self) == 0:
            return ""
        if "TO" not in self:
            # Not valid range, quote
            return "\\%s%s\\%s" % (
                self.start,
                "".join([six.text_type(x) for x in self]),
                self.end,
            )
        else:
            # split on 'TO'
            split = self.index("TO")
            if split > 0:
                first = "".join(
                    [
                        six.text_type(x)
                        for x in self[:split]
                        if not isinstance(x, Whitespace)
                    ]
                )
            if split < (len(self) - 1):
                last = "".join(
                    [
                        six.text_type(x)
                        for x in self[split + 1 :]
                        if not isinstance(x, Whitespace)
                    ]
                )
        return "%s%s TO %s%s" % (self.start, first, last, self.end)


class Stack(list):
    def __init__(self):
        self.append([])

    def add(self, item):
        self.current.append(item)
        self.append(item)

    @property
    def current(self):
        return self[-1]

    def __str__(self):
        return "".join([six.text_type(x) for x in self[0]])


def quote(term, textfield=False, prefix_wildcard=False):
    if isinstance(term, six.binary_type):
        term = term.decode("utf-8")
    stack = Stack()
    tokens = query_tokenizer.findall(term.strip())
    # Counter enables lookahead
    i = 0
    stop = len(tokens)
    while i < stop:
        whitespace, boolean, text, grouping, special = tokens[i]

        if whitespace:
            # Add whitespace if group text, range and group filter on display
            if isinstance(stack.current, Group):
                stack.current.append(Whitespace())
            elif isinstance(stack.current, list):
                # We have whitespace with no grouping, insert group
                new = Group("(", ")")
                new.extend(stack.current)
                new.append(Whitespace())
                stack.current[:] = []
                stack.add(new)

        elif boolean:
            # It's an operator if the following token is the same...
            if i < stop - 1 and boolean == tokens[i + 1][1]:
                # Add operator if we're inside a group
                if isinstance(stack.current, Group):
                    stack.current.append(boolean)
                elif isinstance(stack.current, list):
                    # We have an operator with no grouping, insert group
                    new = Group("(", ")")
                    new.extend(stack.current)
                    new.append(boolean)
                    stack.current[:] = []
                    stack.add(new)
            else:
                stack.current.append(boolean)

        elif grouping:
            # [] (inclusive range), {} (exclusive range), always with TO inside
            # () group
            # "" for quotes
            if grouping == '"':
                if isinstance(stack.current, Quote):
                    # Handle empty double quote
                    if not stack.current:
                        stack.current.end = '\\"'
                    else:
                        stack.current.start = stack.current.end = '"'
                        stack.current.isgroup = True
                    stack.pop()
                else:
                    # Right now this is just a single quote,
                    # we set proper start and end before popping
                    new = Quote(start='\\"', end="")
                    stack.add(new)
            elif isinstance(stack.current, Quote):
                # If we're in a quote, escape and print
                stack.current.append("\\%s" % grouping)
            elif grouping in "[{":
                new = Range(start=grouping, end={"[": "]", "{": "}"}[grouping])
                stack.add(new)
            elif grouping == "(":
                new = Group(start="(", end=")")
                stack.add(new)
            elif grouping in "]})":
                if isinstance(stack.current, Group) and stack.current.end == grouping:
                    stack.current.isgroup = True
                    stack.pop()
                else:
                    stack.current.append("\\%s" % grouping)

        elif text:
            stack.current.append(text)

        elif special:
            if special == "\\":
                # Inspect next to see if it's quoted special or quoted group
                if (i + 1) < stop:
                    _, _, _, g2, s2 = tokens[i + 1]
                    if s2:
                        stack.current.append("%s%s" % (special, s2))
                        # Jump ahead
                        i += 1
                    elif g2:
                        stack.current.append("%s%s" % (special, g2))
                        # Jump ahead
                        i += 1
                    else:
                        # Quote it
                        stack.current.append("\\%s" % special)
                else:
                    # Quote it
                    stack.current.append("\\\\")
            elif isinstance(stack.current, Quote):
                stack.current.append("\\%s" % special)
            elif special in "+-":
                if (i + 1) < stop:
                    _, _, t2, g2, _ = tokens[i + 1]
                    # We allow + and - in front of phrase and text
                    if t2 or g2 == '"':
                        if textfield and i > 0 and tokens[i - 1][2]:
                            # Quote intra-word hyphens, so they are normal text
                            # and not syntax
                            stack.current.append("\\%s" % special)
                        else:
                            stack.current.append(special)
                    else:
                        # Quote it
                        stack.current.append("\\%s" % special)
            elif special in "~^":
                # Fuzzy or proximity is always after a term or phrase, and
                # sometimes before int or float like roam~0.8 or
                # "jakarta apache"~10
                if i > 0:
                    _, _, t0, g0, _ = tokens[i - 1]
                    if t0 or g0 == '"':
                        # Look ahead to check for integer or float

                        if (i + 1) < stop:
                            _, _, t2, _, _ = tokens[i + 1]
                            try:  # float(t2) might fail
                                if t2 and float(t2):
                                    stack.current.append("%s%s" % (special, t2))
                                    # Jump ahead
                                    i += 1
                                else:
                                    stack.current.append(special)
                            except ValueError:
                                stack.current.append(special)
                        else:  # (i+1)<stop
                            stack.current.append(special)
                    else:  # t0 or g0 == '"'
                        stack.current.append("\\%s" % special)
                else:  # i>0
                    stack.current.append("\\%s" % special)
            elif special in "?*":
                # ? and * can not be the first characters of a search
                # if prefix_wildcard is not enabled
                if (
                    (
                        stack.current
                        and not getattr(stack.current[-1], "isgroup", False)
                        and (
                            isinstance(stack.current[-1], six.text_type)
                            and not stack.current[-1] in special
                        )
                    )
                    or (
                        prefix_wildcard
                        and (not stack.current or not stack.current[-1] in special)
                    )
                    or isinstance(stack.current, Range)
                ):
                    stack.current.append(special)
            elif special in "/":
                stack.current.append("\\%s" % special)
            elif isinstance(stack.current, Range):
                stack.current.append(special)
            elif isinstance(stack.current, Group):
                stack.current.append("\\%s" % special)
            elif isinstance(stack.current, list):
                stack.current.append("\\%s" % special)
        i += 1
    return six.text_type(stack)


def quote_iterable_item(term):
    if isinstance(term, (int, float)):
        return str(term)
    quoted = quote(term)
    if not quoted.startswith('"') and not quoted == term:
        quoted = quote('"' + term + '"')
    return quoted



from Products.CMFCore.interfaces import IIndexQueueProcessor
from zope.interface import Interface
from zope.schema import Bool, Float, Int, List, Password, Text, TextLine
from zope.schema.interfaces import IVocabularyFactory

from collective.solr import SolrMessageFactory as _


# If we send too large a number to solr, it can't parse it
MAX_RESULTS_SUPPORTED_BY_SOLR = 1_000_000_000


class ISolrSchema(Interface):
    active = Bool(
        title=_("label_active", default="Active"),
        description=_(
            "help_active",
            default="Check this to enable the Solr integration, i.e. "
            "indexing and searching using the below settings.",
        ),
        default=False,
        required=False,
    )

    host = TextLine(
        title=_("label_host", default="Host"),
        description=_(
            "help_host", default="The host name of the Solr instance to be used."
        ),
    )

    port = Int(
        title=_("label_port", default="Port"),
        description=_("help_port", default="The port of the Solr instance to be used."),
    )

    solr_login = TextLine(
        title=_("login", default="Login"),
        description=_(
            "help_login", default="Authentication login of the SolR instance."
        ),
        required=False,
    )

    solr_password = Password(
        title=_("password", default="Password"),
        description=_(
            "help_password", default="Authentication password of the SolR instance."
        ),
        required=False,
    )

    base = TextLine(
        title=_("label_base", default="Base"),
        description=_(
            "help_base", default="The base prefix of the Solr instance to be used."
        ),
    )

    async_indexing = Bool(
        title=_("label_async", default="Asynchronous indexing"),
        default=False,
        required=False,
        description=_(
            "help_async",
            default="Check to enable asynchronous indexing operations, "
            "which will improve Zope response times in return for "
            "not having the Solr index updated immediately.",
        ),
    )

    auto_commit = Bool(
        title=_("label_auto_commit", default="Automatic commit"),
        default=True,
        required=False,
        description=_(
            "help_auto_commit",
            default="If enabled each index operation will cause a commit "
            "to be sent to Solr, which causes it to update its "
            "index. If you disable this, you need to configure "
            "commit policies on the Solr server side.",
        ),
    )

    commit_within = Int(
        title=_("label_commit_within", default="Commit within"),
        default=0,
        description=_(
            "help_commit_within",
            default="Maximum number of milliseconds after which adds "
            "should be processed by Solr. Defaults to 0, meaning "
            "immediate commits. Enabling this feature implicitly "
            "disables automatic commit and you should configure "
            "commit policies on the Solr server side. Otherwise "
            "large numbers of deletes without adds will not be "
            "processed. This feature requires a Solr 1.4 server.",
        ),
    )

    index_timeout = Float(
        title=_("label_index_timeout", default="Index timeout"),
        description=_(
            "help_index_timeout",
            default="Number of seconds after which an index request will "
            'time out. Set to "0" to disable timeouts.',
        ),
    )

    search_timeout = Float(
        title=_("label_search_timeout", default="Search timeout"),
        description=_(
            "help_search_timeout",
            default="Number of seconds after which a search request will "
            'time out. Set to "0" to disable timeouts.',
        ),
    )

    max_results = Int(
        title=_("label_max_results", default="Maximum search results"),
        description=_(
            "help_max_results",
            default="Specify the maximum number of matches to be returned "
            'when searching. Set to "10000000" or some other '
            "ridiculously large value that is higher than the "
            "possible number of rows that are expected.",
        ),
        default=1_000_000,
        max=MAX_RESULTS_SUPPORTED_BY_SOLR,
    )

    required = List(
        title=_("label_required", default="Required query parameters"),
        description=_(
            "help_required",
            default="Specify required query parameters, one per line. "
            "Searches will only get dispatched to Solr if any "
            "of the listed parameters is present in the query. "
            "Leave empty to dispatch all searches.",
        ),
        value_type=TextLine(),
        default=[],
        missing_value=[],
        required=False,
    )

    search_pattern = Text(
        title=_("label_search_pattern", default="Pattern for simple search queries"),
        description=_(
            "help_search_pattern",
            default="Specify a query pattern used for simple queries "
            "consisting only of words and numbers, i.e. not "
            "using any of Solr's advanced query expressions. "
            "{value} and {base_value} are available in the "
            "pattern and will be replaced by the search word "
            "and the search word stripped of wildcard symbols.",
        ),
    )

    prefix_wildcard = Bool(
        title=_(
            "label_simple_search_prefix_wildcard",
            default="Allow prefix wildcard searches and use them in simple searches",
        ),
        default=False,
        required=False,
    )

    force_simple_search = Bool(
        title=_("label_force_simple_search", default="Force simple search pattern"),
        description=_(
            "help_force_simple_search",
            default="If set all queries in SearchableText will use the "
            '"Pattern for simple search queries". This will remove all special '
            "solr characters and operators from the query text.",
        ),
        default=False,
        required=False,
    )

    allow_complex_search = Bool(
        title=_("allow_complex_search", default="Allow complex search"),
        description=_(
            "help_allow_complex_search",
            default="Allow a complex search in SearchableText by either prefixing the "
            'search text with "solr:" or passing "solr_complex_search" as a field '
            "in the catalog query.",
        ),
        default=False,
        required=False,
    )

    facets = List(
        title=_("label_facets", default="Default search facets"),
        description=_(
            "help_facets",
            default="Specify catalog indexes that should be queried for "
            "facet information, one per line.",
        ),
        value_type=TextLine(),
        default=[],
        required=False,
    )

    filter_queries = List(
        title=_("label_filter_queries", default="Filter query parameters"),
        description=_(
            "help_filter_queries",
            default="Specify query parameters for which filter queries "
            "should be used, one per line.  You can use several "
            "indices in one filter query separated by space. "
            "Typical examples are "
            '"effective expires allowedRolesAndUsers" or '
            '"review_state portal_type".',
        ),
        value_type=TextLine(),
        default=[],
        required=False,
    )

    slow_query_threshold = Int(
        title=_("label_slow_query_threshold", default="Slow query threshold"),
        description=_(
            "help_slow_query_threshold",
            default="Specify a threshold (in milliseconds) after which "
            "queries are considered to be slow causing them to "
            'be logged. Set to "0" to prevent any logging.',
        ),
        default=0,
    )

    effective_steps = Int(
        title=_("label_effective_steps", default="Effective date steps"),
        default=1,
        description=_(
            "help_effective_steps",
            default="Specify the effective date steps in seconds. "
            "Using 900 seconds (15 minutes) means the effective "
            "date sent to Solr changes every 15 minutes.",
        ),
    )

    exclude_user = Bool(
        title=_("label_exclude_user", default="Exclude user from allowedRolesAndUsers"),
        description=_(
            "help_exclude_user",
            default="Specify whether the user:userid should be excluded "
            "from allowedRolesAndUsers to improve cacheability "
            "on the expense of finding content with local roles"
            "given to specific users.",
        ),
        default=False,
        required=False,
    )

    highlight_fields = List(
        title=_("label_highlight_fields", "Highlighting fields"),
        description=_(
            "help_highlight_fields",
            default=(
                "Fields that should be used for highlighting. "
                "Snippets of text will be generated from the contents "
                "of these fields, with the search keywords that "
                "matched highlighted inside the text."
            ),
        ),
        value_type=TextLine(),
        default=[],
        required=False,
    )

    highlight_formatter_pre = TextLine(
        title=_("label_highlight_formatter_pre", default="Highlight formatter: pre"),
        description=_(
            "help_highlight_formatter_pre",
            default="The text to insert before the highlighted keyword.",
        ),
        default="[",
        required=False,
    )

    highlight_formatter_post = TextLine(
        title=_("label_highlight_formatter_post", default="Highlight formatter: post"),
        description=_(
            "help_highlight_formatter_post",
            default="The text to insert after the highlighted keyword.",
        ),
        default="]",
        required=False,
    )

    highlight_fragsize = Int(
        title=_("label_highlight_fragsize", default="Highlight Fragment Size"),
        description=_(
            "help_highlight_fragsize",
            default=(
                "The size, in characters, of the snippets (aka "
                "fragments) created by the highlighter."
            ),
        ),
        default=100,
    )

    field_list = List(
        title=_("label_field_list", default="Default fields to be returned"),
        description=_(
            "help_field_list",
            default=(
                "Specify metadata fields that should be returned for "
                "items in the result set, one per line. Defaults to "
                "all available plus ranking score."
            ),
        ),
        value_type=TextLine(),
        default=[],
        required=False,
    )

    levenshtein_distance = Float(
        title=_("label_levenshtein_distance", default="Levenshtein distance"),
        description=_(
            "help_levenshtein_distance",
            default="The Levenshtein distance is a string metric for "
            "measuring the difference between two strings. It allows"
            "you to perform fuzzy searches by specifying a value "
            "between 0 and 1.",
        ),
        required=False,
        default=0.0,
    )

    atomic_updates = Bool(
        title=_("label_atomic_updates", default="Enable atomic updates"),
        description=_(
            "help_atomic_updates",
            default="Atomic updates allows you to update only specific "
            'indexes, like "reindexObject(idxs=["portal_type"])".'
            "Unfortunately atomic updates are not compatible with "
            "index time boosting. If you enable atomic updates, "
            "index time boosting no longer works.",
        ),
        default=True,
        required=False,
    )

    boost_script = Text(
        title=_(
            "label_boost_script", default="Python script for custom index boosting"
        ),
        required=False,
        default="",
        missing_value="",
        description=_(
            "help_boost_script",
            default="This script is meant to be customized according to "
            "site-specific search requirements, e.g. boosting "
            'certain content types like "news items", ranking older '
            "content lower, consider special important content items,"
            " content rating etc."
            " the indexing data that will be sent to Solr is passed "
            "in as the `data` parameter, the indexable object is "
            "available via the `context` binding. The return value "
            "should be a dictionary consisting of field names and "
            "their respecitive boost values.  use an empty string "
            "as the key to set a boost value for the entire "
            "document/content item.",
        ),
    )

    use_tika = Bool(
        title=_("label_use_tika", default="Use Tika"),
        description=_(
            "help_use_tika",
            default="Upload binary files to Solr via Tika. "
            "That way Solr does not need direct access to the blob files on the file system. "
            "Use this setting when Solr runs on a separate server or if you use Relstorage instead of ZEO.",
        ),
        default=False,
        required=False,
    )

    tika_default_field = TextLine(
        title=_("label_tika_default_field", default="Tika default field"),
        description=_(
            "help_tika_default_field",
            default="Field that Tika uses to add the extracted text to.",
        ),
        default="content",
        required=False,
    )

    stopwords = Text(
        title=_("label_stopwords", default="Stopwords in the format of stopwords.txt"),
        description=_(
            "help_stopwords",
            default="Copy the stopwords.txt file here. "
            "Check Solr configuration to understand the format. - "
            "Stopwords will not get (word OR word*) simple "
            "expression, only (word). "
            "Notes: "
            "1. This will only work for multi word queries "
            "when force_simple_expression=True. - "
            "2. It's still necessary to filter stopwords from "
            "Solr, this option only causes the "
            "faulty (stopword*) parts removed from "
            "the expression ",
        ),
        default="",
        required=False,
    )

    stopwords_case_insensitive = Bool(
        title=_(
            "label_stopwords_case_insensitive", default="Stopwords are case insensitive"
        ),
        description=_(
            "help_stopwords_are_case_insensitive",
            default="Stopwords are case insensitive "
            "This depends on your Solr setup. If your stopwords are processed in a case insensitive way, "
            "this should be checked and it will apply the stopword wildcard removal in a case "
            "insensitive way.",
        ),
        default=False,
        required=False,
    )


class ISolrConnectionConfig(ISolrSchema):
    """utility to hold the connection configuration for the solr server"""


class IZCMLSolrConnectionConfig(Interface):
    """Solr connection settings configured through ZCML."""


class ISolrConnectionManager(Interface):
    """a thread-local connection manager for solr"""

    def setHost(active=False, host="localhost", port=8983, base="/solr/plone"):
        """set connection parameters"""

    def closeConnection(clearSchema=False):
        """close the current connection, if any"""

    def getConnection():
        """returns an existing connection or opens one"""

    def getSchema():
        """returns the currently used schema or fetches it.
        If the schema cannot be fetched None is returned."""

    def setTimeout(timeout, lock=object()):
        """set the timeout on the current (or to be opened) connection
        to the given value and optionally lock it until explicitly
        freed again"""

    def setIndexTimeout():
        """set the timeout on the current (or to be opened) connection
        to the value specified for indexing operations"""

    def setSearchTimeout():
        """set the timeout on the current (or to be opened) connection
        to the value specified for search operations"""


class ISolrIndexQueueProcessor(IIndexQueueProcessor):
    """an index queue processor for solr"""


class ISolrFlare(Interface):
    """a sol(a)r brain, i.e. a data container for search results"""


class IFlare(Interface):
    """marker interface for pluggable brain wrapper classes, providing
    additional helper methods like `getURL` etc"""


class ISearch(Interface):
    """a generic search interface
    FIXME: this should be defined in a generic package"""

    def search(query, **parameters):
        """perform a search with the given querystring and extra parameters
        (see http://wiki.apache.org/solr/CommonQueryParameters)"""

    def __call__(query, **parameters):
        """convenience alias for `search`"""

    def buildQueryAndParameters(default=None, **args):
        """helper to build a query for simple use-cases; the query is
        returned as a dictionary which might be string-joined or
        passed to the `search` method as the `query` argument,
        additionally search parameters are substracted from the
        args and returned as a separate dict"""


class ICatalogTool(Interface):
    """marker interface for plone's catalog tool"""


class ISearchDispatcher(Interface):
    """adapter for potentially dispatching a given query to an
    alternative search backend (instead of the portal catalog)"""

    def __call__(request, **keywords):
        """decide if an alternative search backend is capable of performing
        the given query and use it or fall back to the portal catalog"""


class ISolrMaintenanceView(Interface):
    """solr maintenance view for clearing, re-indexing content etc"""

    def optimize():
        """optimize solr indexes"""

    def clear():
        """clear all data from solr, i.e. delete all indexed objects"""

    def reindex(batch=1000, skip=0):
        """find all contentish objects (meaning all objects derived from one
        of the catalog mixin classes) and (re)indexes them"""

    def sync(batch=1000):
        """sync the solr index with the portal catalog;  records contained
        in the catalog but not in solr will be indexed and records not
        contained in the catalog can be optionally removed;  this can
        be used to ensure consistency between zope and solr after the
        solr server has been unavailable etc"""

    def cleanup(batch=1000):
        """remove entries from solr that don't have a corresponding Zope
        object  or have a different UID than the real object"""


class ISolrAddHandler(Interface):
    """An adder for solr documents"""


class IFacetTitleVocabularyFactory(IVocabularyFactory):
    """A vocabulary factory used to create a vocabulary that provides titles
    for facet values

    When facet values are displayed for selection on the search results page, a
    named IFacetTitleVocabularyFactory is looked up, and if it exists it's
    called to return a zope.schema.IBaseVocabulary vocabulary. The name is the
    same as the facet name (e.g. "portal_type" or "review_state"). This
    vocabulary should return zope.schema.ITitledTokenizedTerm items, their
    title attribute is what is displayed in the UI.
    """


class ICheckIndexable(Interface):
    """Check if an object is indexable"""

    def __call__():
        """Return `True`, if context is indexable and `False`otherwise"""



from unittest import TestCase

from collective.solr.contentlisting import FlareContentListingObject
from collective.solr.flare import PloneFlare
from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING
from DateTime import DateTime
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.uuid.interfaces import IUUID
from zope.interface.verify import verifyClass


class ContentListingTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.flare = FlareContentListingObject(
            PloneFlare(
                {
                    "getId": "foobar",
                    "path_string": "/plone/news",
                    "UID": "test-uid",
                    "getObjSize": 42,
                    "review_state": "published",
                    "listCreators": ["foo", "bar"],
                    "Creator": "Flare Creator",
                    "Title": "Flare Title",
                    "Description": "Flare Description",
                    "Subject": "Flare Subject",
                    "Date": "Flare Date",
                    "expires": DateTime("1.1.2099"),
                    "created": DateTime("31.12.1969"),
                    "modified": DateTime("27.07.2016"),
                    "Language": "de",
                    "portal_type": "NewsItem",
                    "Type": "Flare NewsItem",
                    "othermetadata": "Flare Othermetadata",
                }
            )
        )

    def testInterfaceComplete(self):
        self.assertTrue(verifyClass(IContentListingObject, FlareContentListingObject))

    def test_getId(self):
        self.assertEqual(self.flare.getId(), "foobar")

    def test_getObject(self):
        self.assertEqual(self.flare.getObject(), self.layer["portal"]["news"])

    def test_getDataOrigin(self):
        self.assertEqual(self.flare.getObject(), self.layer["portal"]["news"])

    def test_getPath(self):
        self.assertEqual(self.flare.getPath(), "/plone/news")

    def test_getURL(self):
        self.assertEqual(self.flare.getURL(False), "/plone/news")
        self.assertEqual(self.flare.getURL(True), "/plone/news")

    def test_uuid_key(self):
        self.assertEqual(self.flare.uuid(), "test-uid")

    def test_uuid_object(self):
        del self.flare.flare["UID"]
        self.assertEqual(self.flare.uuid(), IUUID(self.layer["portal"]["news"]))

    def test_getSize(self):
        self.assertEqual(self.flare.getSize(), 42)

    def test_review_state(self):
        self.assertEqual(self.flare.review_state(), "published")

    def test_listCreators(self):
        self.assertEqual(self.flare.listCreators(), ["foo", "bar"])

    def test_Creator(self):
        self.assertEqual(self.flare.Creator(), "Flare Creator")

    def test_Subject(self):
        self.assertEqual(self.flare.Subject(), "Flare Subject")

    def test_Publisher(self):
        self.assertRaises(NotImplementedError, self.flare.Publisher)

    def test_listContributors(self):
        self.assertRaises(NotImplementedError, self.flare.listContributors)

    def test_Contributors(self):
        self.assertRaises(NotImplementedError, self.flare.Contributors)

    def test_Date(self):
        self.assertEqual(self.flare.Date(), "Flare Date")

    def test_CreationDate(self):
        self.assertEqual(self.flare.CreationDate().ISO(), "1969-12-31T00:00:00")

    def test_EffectiveDate(self):
        self.assertEqual(
            self.flare.EffectiveDate(), self.layer["portal"]["news"].EffectiveDate()
        )

    def test_ExpirationDate(self):
        self.assertEqual(self.flare.ExpirationDate().ISO(), "2099-01-01T00:00:00")

    def test_ModificationDate(self):
        self.assertEqual(self.flare.ModificationDate().ISO(), "2016-07-27T00:00:00")

    def test_Format(self):
        self.assertRaises(NotImplementedError, self.flare.Format)

    def test_Identifier(self):
        self.assertEqual(self.flare.Identifier(), "/plone/news")

    def test_Language(self):
        self.assertEqual(self.flare.Language(), "de")

    def test_Rights(self):
        self.assertRaises(NotImplementedError, self.flare.Rights)

    def test_Title(self):
        self.assertEqual(self.flare.Title(), "Flare Title")

    def test_Description(self):
        self.assertEqual(self.flare.Description(), "Flare Description")

    def test_Type(self):
        self.assertEqual(self.flare.Type(), "Flare NewsItem")

    def test_ContentTypeClass(self):
        self.assertEqual(self.flare.ContentTypeClass(), "contenttype-newsitem")

    def test_PortalType(self):
        self.assertEqual(self.flare.PortalType(), "NewsItem")

    def test_Author(self):
        self.assertEqual(
            self.flare.Author(),
            {
                "username": "Flare Creator",
                "description": "",
                "language": "",
                "home_page": "/HOMEPAGEURL",
                "location": "",
                "fullname": "Flare Creator",
            },
        )

    def test_CroppedDescription(self):
        self.assertEqual(self.flare.CroppedDescription(), "Flare Description")

    def test_pretty_title(self):
        self.assertEqual(self.flare.flare.pretty_title_or_id(), "Flare Title")

    def test_creation_date(self):
        self.assertTrue(self.flare.flare.CreationDate.startswith("1969-12-31T"))

    def test_UID(self):
        self.assertEqual(self.flare.UID, self.flare.flare.UID)

    def test_portal_type(self):
        self.assertEqual(self.flare.portal_type, self.flare.PortalType())

    def test_modified(self):
        self.assertEqual(self.flare.modified, self.flare.ModificationDate())

    def test_defer_lookup_to_flare(self):
        self.assertEqual(self.flare.othermetadata, "Flare Othermetadata")



import unittest

from collective.solr.contentlisting import FlareContentListingObject
from collective.solr.testing import (
    LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING,
    activateAndReindex,
)
from plone.app.testing import TEST_USER_ID, setRoles


class CollectionsTestCase(unittest.TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING

    def setUp(self):
        portal = self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        id_ = portal.invokeFactory("Collection", id="foo", title="Foo")
        collection = self.collection = portal[id_]
        self.query = query = [
            {
                "i": "SearchableText",
                "o": "plone.app.querystring.operation.string.contains",
                "v": "new",
            }
        ]
        collection.query = query
        activateAndReindex(portal)

    def test_has_results(self):
        # Make sure we get some results
        # If not, then the other tests won't do anything.
        results = self.collection.results()
        self.assertGreater(results.length, 0)
        self.assertTrue(isinstance(results[0], FlareContentListingObject))

    def test_render_collection(self):
        # If it renders, it's good.
        self.collection()

    def test_render_querybuilder_html_results(self):
        # If it renders, it's good.
        collection = self.collection
        view = collection.unrestrictedTraverse("@@querybuilder_html_results")
        view.html_results(self.query)



from datetime import date, datetime
from re import DOTALL, findall, search
from threading import Thread
from unittest import TestCase

from collective.solr.indexer import SolrIndexProcessor
from collective.solr.indexer import logger as logger_indexer
from collective.solr.interfaces import ICheckIndexable
from collective.solr.manager import SolrConnectionManager
from collective.solr.solr import SolrConnection
from collective.solr.testing import COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE
from collective.solr.tests.utils import fakehttp, fakemore, getData
from collective.solr.utils import getConfig, prepareData
from persistent.mapping import PersistentMapping
from DateTime import DateTime
from Products.CMFCore.CMFCatalogAware import CMFCatalogAware
from zope.interface import implementer


@implementer(ICheckIndexable)
class Foo(CMFCatalogAware):

    """dummy test object"""

    def __init__(self, **kw):
        for key, value in kw.items():
            setattr(self, key, value)

    def __call__(self):
        return True


def sortFields(output):
    """helper to sort `<field>` tags in output for testing"""
    pattern = rb"^(.*<doc>)(<field .*</field>)(</doc>.*)"
    prefix, fields, suffix = search(pattern, output, DOTALL).groups()
    tags = rb"(<field [^>]*>[^<]*</field>)"
    return prefix + b"".join(sorted(findall(tags, fields))) + suffix


class QueueIndexerTests(TestCase):

    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.mngr = SolrConnectionManager()
        self.mngr.setHost(active=True)
        conn = self.mngr.getConnection()
        fakehttp(conn, getData("schema.xml"))  # fake schema response
        self.mngr.getSchema()  # read and cache the schema
        self.proc = SolrIndexProcessor(self.mngr)
        config = getConfig()
        config.atomic_updates = True

    def tearDown(self):
        self.mngr.closeConnection()
        self.mngr.setHost(active=False)

    def testPrepareData(self):
        data = {"allowedRolesAndUsers": ["user:test_user_1_", "user:portal_owner"]}
        prepareData(data)
        self.assertEqual(
            data, {"allowedRolesAndUsers": ["user$test_user_1_", "user$portal_owner"]}
        )

    def testLanguageParameterHandling(self):
        # empty strings are replaced...
        data = {"Language": ["en", ""]}
        prepareData(data)
        self.assertEqual(data, {"Language": ["en", "any"]})
        data = {"Language": ""}
        prepareData(data)
        self.assertEqual(data, {"Language": "any"})
        # for other indices this shouldn't happen...
        data = {"Foo": ["en", ""]}
        prepareData(data)
        self.assertEqual(data, {"Foo": ["en", ""]})

    def testIndexObject(self):
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        # indexing sends data
        self.proc.index(Foo(id="500", name="python test doc"))
        self.assertEqual(
            sortFields(str(output).encode("utf-8")),
            getData("add_request.txt").rstrip(b"\n"),
        )

    def testIndexAccessorRaises(self):
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)

        def brokenfunc():
            raise ValueError

        self.proc.index(
            Foo(id="500", name="python test doc", text=brokenfunc)
        )  # indexing sends data
        self.assertEqual(
            sortFields(str(output).encode("utf-8")),
            getData("add_request.txt").rstrip(b"\n"),
        )

    def testPartialIndexObject(self):
        foo = Foo(id="500", name="foo", price=42.0)
        # first index all attributes...
        response = getData("add_response.txt")
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        self.assert_(
            str(output).find('<field name="price" update="set">42.0</field>') > 0,
            '"price" data not found',
        )
        # then only a subset...
        response = getData("add_response.txt")
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo, attributes=["id", "name"])
        output = str(output)
        self.assert_(
            output.find('<field name="name" update="set">foo</field>') > 0,
            '"name" data not found',
        )
        # at this point we'd normally check for a partial update:
        self.assertEqual(output.find("price"), -1, '"price" data found?')
        self.assertEqual(output.find("42"), -1, '"price" data found?')

    def testDateIndexing(self):
        foo = Foo(
            id="zeidler",
            name="andi",
            cat="nerd",
            timestamp=DateTime("May 11 1972 03:45:59.999730 GMT"),
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required = (
            '<field name="timestamp" update="set">' "1972-05-11T03:45:59.999Z</field>"
        )
        self.assert_(str(output).find(required) > 0, '"date" data not found')

    def testDateIndexingWithPythonDateTime(self):
        foo = Foo(
            id="gerken",
            name="patrick",
            cat="nerd",
            timestamp=datetime(1980, 9, 29, 14, 0o2, 59, 999730),
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required = (
            '<field name="timestamp" update="set">' "1980-09-29T14:02:59.999Z</field>"
        )
        self.assert_(str(output).find(required) > 0, '"date" data not found')

    def testDateIndexingWithPythonDate(self):
        foo = Foo(
            id="brand", name="jan-carel", cat="nerd", timestamp=date(1982, 8, 0o5)
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required = (
            '<field name="timestamp" update="set">' "1982-08-05T00:00:00.000Z</field>"
        )
        self.assert_(str(output).find(required) > 0, '"date" data not found')

    def testReindexObject(self):
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        # reindexing sends data
        self.proc.reindex(Foo(id="500", name="python test doc"))
        self.assertEqual(
            sortFields(str(output).encode("utf-8")),
            getData("add_request.txt").rstrip(b"\n"),
        )

    def testUnindexObject(self):
        response = getData("delete_response.txt")
        # fake response
        output = fakehttp(self.mngr.getConnection(), response)
        # unindexing sends data
        self.proc.unindex(Foo(id="500", name="python test doc"))
        self.assertEqual(
            str(output), getData("delete_request.txt").decode("utf-8").rstrip("\n")
        )

    def testCommit(self):
        response = getData("commit_response.txt")
        # fake response
        output = fakehttp(self.mngr.getConnection(), response)
        # committing sends data
        self.proc.commit()
        self.assertEqual(
            str(output), getData("commit_request.txt").decode("utf-8").rstrip("\n")
        )

    def testNoIndexingWithoutAllRequiredFields(self):
        response = getData("dummy_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        # indexing sends data
        self.proc.index(Foo(id="500"))
        self.assertEqual(str(output), "")

    def testIndexerMethods(self):
        class Bar(Foo):
            def cat(self):
                return "nerd"

            def price(self):
                raise AttributeError("price")

        foo = Bar(id="500", name="foo")
        # raising the exception should keep the attribute from being indexed
        response = getData("add_response.txt")
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        output = str(output)
        self.assertTrue(
            output.find('<field name="cat" update="set">nerd</field>') > 0,
            '"cat" data not found',
        )
        self.assertEqual(output.find("price"), -1, '"price" data found?')

    def testImageScalesIndexing(self):
        foo = Foo(
            id="image",
            name="image",
            image_scales=PersistentMapping({"key": "value"}),
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required = '<field name="image_scales" update="set">{"key": "value"}</field>'
        self.assert_(str(output).find(required) > 0, '"image_scales" data not found')

    def testListIndexing(self):
        foo = Foo(
            id="zeidler",
            name="andi",
            cat=["rock", "paper", "Sharp scissors"],
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required1 = '<field name="cat" update="set">rock</field>'
        required2 = '<field name="cat" update="set">paper</field>'
        required3 = '<field name="cat" update="set">Sharp scissors</field>'
        self.assert_(str(output).find(required1) > 0, "Some element (rock) not found")
        self.assert_(str(output).find(required2) > 0, "Some element (paper) not found")
        self.assert_(
            str(output).find(required3) > 0, "Some element (Sharp scissors) not found"
        )

    def testTupleIndexing(self):
        foo = Foo(
            id="zeidler",
            name="andi",
            cat=("rock", "paper", "Sharp scissors"),
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required1 = '<field name="cat" update="set">rock</field>'
        required2 = '<field name="cat" update="set">paper</field>'
        required3 = '<field name="cat" update="set">Sharp scissors</field>'
        self.assert_(str(output).find(required1) > 0, "Some element (rock) not found")
        self.assert_(str(output).find(required2) > 0, "Some element (paper) not found")
        self.assert_(
            str(output).find(required3) > 0, "Some element (Sharp scissors) not found"
        )

    def testSetIndexing(self):
        foo = Foo(
            id="zeidler",
            name="andi",
            cat={"rock", "paper", "Sharp scissors"},
        )
        response = getData("add_response.txt")
        # fake add response
        output = fakehttp(self.mngr.getConnection(), response)
        self.proc.index(foo)
        required1 = '<field name="cat" update="set">rock</field>'
        required2 = '<field name="cat" update="set">paper</field>'
        required3 = '<field name="cat" update="set">Sharp scissors</field>'
        self.assert_(str(output).find(required1) > 0, "Some element (rock) not found")
        self.assert_(str(output).find(required2) > 0, "Some element (paper) not found")
        self.assert_(
            str(output).find(required3) > 0, "Some element (Sharp scissors) not found"
        )


class RobustnessTests(TestCase):

    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.mngr = SolrConnectionManager()
        self.mngr.setHost(active=True)
        self.conn = self.mngr.getConnection()
        self.proc = SolrIndexProcessor(self.mngr)
        self.log = []  # catch log messages...

        def logger(*args):
            self.log.extend(args)

        logger_indexer.warning = logger
        config = getConfig()
        config.atomic_updates = True

    def tearDown(self):
        self.mngr.closeConnection()
        self.mngr.setHost(active=False)

    def testIndexingWithUniqueKeyMissing(self):
        # fake schema response
        fakehttp(self.conn, getData("simple_schema.xml"))
        # read and cache the schema
        self.mngr.getSchema()
        response = getData("add_response.txt")
        output = fakehttp(self.conn, response)  # fake add response
        foo = Foo(id="500", name="foo")
        # indexing sends data
        self.proc.index(foo)
        # nothing happened...
        self.assertEqual(len(output), 0)
        self.assertEqual(
            self.log, ["schema is missing unique key, skipping indexing of %r", foo]
        )

    def testUnindexingWithUniqueKeyMissing(self):
        # fake schema response
        fakehttp(self.conn, getData("simple_schema.xml"))
        # read and cache the schema
        self.mngr.getSchema()
        response = getData("delete_response.txt")
        # fake delete response
        output = fakehttp(self.conn, response)
        foo = Foo(id="500", name="foo")
        # unindexing sends data
        self.proc.unindex(foo)
        # nothing happened...
        self.assertEqual(len(output), 0)
        self.assertEqual(
            self.log, ["schema is missing unique key, skipping unindexing of %r", foo]
        )


class FakeHTTPConnectionTests(TestCase):

    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.foo = Foo(id="500", name="python test doc")
        self.schema_request = "GET /solr/plone/admin/file/?file=schema.xml"
        config = getConfig()
        config.atomic_updates = True

    def testSingleRequest(self):
        mngr = SolrConnectionManager(active=True)
        output = fakehttp(mngr.getConnection(), getData("schema.xml"))
        mngr.getSchema()
        mngr.closeConnection()
        self.failUnless(output.get().decode("utf-8").startswith(self.schema_request))

    def testTwoRequests(self):
        mngr = SolrConnectionManager(active=True)
        proc = SolrIndexProcessor(mngr)
        output = fakehttp(
            mngr.getConnection(), getData("schema.xml"), getData("add_response.txt")
        )
        proc.index(self.foo)
        mngr.closeConnection()
        self.assertEqual(len(output), 2)
        self.failUnless(output.get().decode("utf-8").startswith(self.schema_request))
        self.assertEqual(
            sortFields(output.get()), getData("add_request.txt").rstrip(b"\n")
        )

    def testThreeRequests(self):
        mngr = SolrConnectionManager(active=True)
        proc = SolrIndexProcessor(mngr)
        output = fakehttp(
            mngr.getConnection(),
            getData("schema.xml"),
            getData("add_response.txt"),
            getData("delete_response.txt"),
        )
        proc.index(self.foo)
        proc.unindex(self.foo)
        mngr.closeConnection()
        self.assertEqual(len(output), 3)
        self.failUnless(output.get().decode("utf-8").startswith(self.schema_request))
        self.assertEqual(
            sortFields(output.get()), getData("add_request.txt").rstrip(b"\n")
        )
        self.assertEqual(output.get(), getData("delete_request.txt").rstrip(b"\n"))

    def testFourRequests(self):
        mngr = SolrConnectionManager(active=True)
        proc = SolrIndexProcessor(mngr)
        output = fakehttp(
            mngr.getConnection(),
            getData("schema.xml"),
            getData("add_response.txt"),
            getData("delete_response.txt"),
            getData("commit_response.txt"),
        )
        proc.index(self.foo)
        proc.unindex(self.foo)
        proc.commit()
        mngr.closeConnection()
        self.assertEqual(len(output), 4)
        self.failUnless(output.get().decode("utf-8").startswith(self.schema_request))
        self.assertEqual(
            sortFields(output.get()), getData("add_request.txt").rstrip(b"\n")
        )
        self.assertEqual(output.get(), getData("delete_request.txt").rstrip(b"\n"))
        self.assertEqual(output.get(), getData("commit_request.txt").rstrip(b"\n"))

    def testExtraRequest(self):
        # basically the same as `testThreeRequests`, except it
        # tests adding fake responses consecutively
        mngr = SolrConnectionManager(active=True)
        proc = SolrIndexProcessor(mngr)
        conn = mngr.getConnection()
        output = fakehttp(conn, getData("schema.xml"))
        fakemore(conn, getData("add_response.txt"))
        proc.index(self.foo)
        fakemore(conn, getData("delete_response.txt"))
        proc.unindex(self.foo)
        mngr.closeConnection()
        self.assertEqual(len(output), 3)
        self.failUnless(output.get().decode("utf-8").startswith(self.schema_request))
        self.assertEqual(
            sortFields(output.get()), getData("add_request.txt").rstrip(b"\n")
        )
        self.assertEqual(output.get(), getData("delete_request.txt").rstrip(b"\n"))


class ThreadedConnectionTests(TestCase):

    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def testLocalConnections(self):
        config = getConfig()
        config.atomic_updates = True
        mngr = SolrConnectionManager(active=True)
        proc = SolrIndexProcessor(mngr)
        mngr.setHost(active=True)
        schema = getData("schema.xml")
        log = []

        def runner():
            # fake schema response on solr connection - caches the schema
            fakehttp(mngr.getConnection(), getData("schema.xml"))
            mngr.getConnection().get_schema()

            fakehttp(mngr.getConnection(), schema)  # fake schema response
            # read and cache the schema
            mngr.getSchema()
            response = getData("add_response.txt")
            # fake add response
            output = fakehttp(mngr.getConnection(), response)
            # indexing sends data
            proc.index(Foo(id="500", name="python test doc"))
            mngr.closeConnection()
            log.append(str(output))
            log.append(proc)
            log.append(mngr.getConnection())

        # after the runner was set up, another thread can be created and
        # started;  its output should contain the proper indexing request,
        # whereas the main thread's connection remain idle;  the latter
        # cannot be checked directly, but the connection object would raise
        # an exception if it was used to send a request without setting up
        # a fake response beforehand...
        thread = Thread(target=runner)
        thread.start()
        thread.join()
        conn = mngr.getConnection()  # get this thread's connection
        fakehttp(conn, schema)  # fake schema response
        mngr.getSchema()  # read and cache the schema
        mngr.closeConnection()
        mngr.setHost(active=False)
        self.assertEqual(len(log), 3)
        self.assertEqual(
            sortFields(log[0].encode("utf-8")), getData("add_request.txt").rstrip(b"\n")
        )
        self.failUnless(isinstance(log[1], SolrIndexProcessor))
        self.failUnless(isinstance(log[2], SolrConnection))
        self.failUnless(isinstance(proc, SolrIndexProcessor))
        self.failUnless(isinstance(conn, SolrConnection))
        self.assertEqual(log[1], proc)  # processors should be the same...
        self.assertNotEqual(log[2], conn)  # but not the connections



import os
import subprocess
import sys
from time import sleep

import six
import six.moves.urllib.error
import six.moves.urllib.parse
import six.moves.urllib.request
from collective.solr.utils import activate
from plone.api.portal import set_registry_record
from plone.app.contenttypes.testing import (
    PLONE_APP_CONTENTTYPES_FIXTURE as PLONE_FIXTURE,
)
from plone.app.robotframework.testing import REMOTE_LIBRARY_BUNDLE_FIXTURE
from plone.app.testing import (
    TEST_USER_NAME,
    FunctionalTesting,
    IntegrationTesting,
    PloneSandboxLayer,
    applyProfile,
    login,
)
from plone.registry.interfaces import IRegistry
from plone.testing import Layer, z2
from Products.CMFCore.utils import getToolByName
from six.moves import range
from zope.component import provideUtility
from zope.interface import implementer
from ZPublisher.HTTPRequest import HTTPRequest

BIN_DIR = os.path.dirname(os.path.realpath(sys.argv[0]))


class SolrLayer(Layer):
    """Solr test layer that fires up and shuts down a Solr instance. This
    layer can be used to unit test a Solr configuration without having to
    fire up Plone.
    """

    proc = None

    def __init__(
        self,
        bases=None,
        name="Solr Layer",
        module=None,
        solr_host="localhost",
        solr_port=8983,
        solr_base="/solr/plone",
    ):
        super(SolrLayer, self).__init__(bases, name, module)
        self.solr_host = solr_host
        self.solr_port = solr_port
        self.solr_base = solr_base
        self.solr_url = "http://{0}:{1}{2}".format(solr_host, solr_port, solr_base)

    def setUp(self):
        """Start Solr and poll until it is up and running."""
        self.proc = subprocess.call(
            "./solr-start", shell=True, close_fds=True, cwd=BIN_DIR
        )
        # Poll Solr until it is up and running
        solr_ping_url = "{0}/admin/ping?wt=xml".format(self.solr_url)
        for i in range(1, 10):
            try:
                result = six.moves.urllib.request.urlopen(solr_ping_url)
                if result.code == 200:
                    if b'<str name="status">OK</str>' in result.read():
                        break
            except six.moves.urllib.error.URLError:
                sleep(3)
                sys.stdout.write(".")
            if i == 9:
                subprocess.call("./solr-stop", shell=True, close_fds=True, cwd=BIN_DIR)
                sys.stdout.write("Solr Instance could not be started !!!")

    def tearDown(self):
        """Stop Solr."""
        subprocess.check_call("./solr-stop", shell=True, close_fds=True, cwd=BIN_DIR)


SOLR_FIXTURE = SolrLayer()


class CollectiveSolrLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE,)

    def __init__(
        self,
        bases=None,
        name="Collective Solr Layer",
        module=None,
        solr_host="localhost",
        solr_port=8983,
        solr_base="/solr/plone",
        solr_active=False,
    ):
        super(PloneSandboxLayer, self).__init__(bases, name, module)
        self.solr_active = solr_active
        self.solr_host = solr_host
        self.solr_port = solr_port
        self.solr_base = solr_base
        # SolrLayer should use the same settings as CollectiveSolrLayer
        self.solr_layer = SolrLayer(
            bases,
            name,
            module,
            solr_host=solr_host,
            solr_port=solr_port,
            solr_base=solr_base,
        )

    def setUpZope(self, app, configurationContext):
        # In a full instance the retry_max_count is set in
        # Zope2.startup.handlers. This does not happen in tests, but we need a
        # non-zero retry count for test_retry_on_conflict.
        self._orig_retry_max_count = HTTPRequest.retry_max_count
        HTTPRequest.retry_max_count = 3

        import collective.solr

        self.loadZCML(package=collective.solr)

    def tearDownZope(self, app):
        HTTPRequest.retry_max_count = self._orig_retry_max_count

    def setUpPloneSite(self, portal):
        self.solr_layer.setUp()
        applyProfile(portal, "collective.solr:default")
        set_registry_record("collective.solr.active", self.solr_active)
        set_registry_record("collective.solr.port", self.solr_port)
        set_registry_record("collective.solr.base", self.solr_base)

    def tearDownPloneSite(self, portal):
        set_registry_record("collective.solr.active", False)
        set_registry_record("collective.solr.port", 8983)
        set_registry_record("collective.solr.base", "/solr/plone")
        self.solr_layer.tearDown()


class LegacyCollectiveSolrLayer(CollectiveSolrLayer):
    def setUpPloneSite(self, portal):
        super(LegacyCollectiveSolrLayer, self).setUpPloneSite(portal)
        acl_users = getToolByName(portal, "acl_users")
        acl_users.userFolderAddUser("user1", "secret", ["Manager"], [])
        login(portal, "user1")
        portal.portal_workflow.setDefaultChain("simple_publication_workflow")
        wfAction = portal.portal_workflow.doActionFor
        portal.invokeFactory("Document", id="front-page", title="Welcome to Plone")
        portal.invokeFactory("Folder", id="events", title="EventsFolder")
        portal.invokeFactory("Folder", id="news", title="NewsFolder")
        portal.news.invokeFactory("Collection", id="aggregator", title="News")
        portal.events.invokeFactory("Collection", id="aggregator", title="Events")
        wfAction(portal["front-page"], "publish")
        wfAction(portal.events, "publish")
        wfAction(portal.news, "publish")
        wfAction(portal.news.aggregator, "publish")
        wfAction(portal.events.aggregator, "publish")
        login(portal, TEST_USER_NAME)


LEGACY_COLLECTIVE_SOLR_FIXTURE = LegacyCollectiveSolrLayer()


def activateAndReindex(portal):
    """activate solr indexing and reindex the existing content"""
    activate()
    response = portal.REQUEST.RESPONSE
    original = response.write
    response.write = lambda x: x  # temporarily ignore output
    maintenance = portal.unrestrictedTraverse("@@solr-maintenance")
    maintenance.clear()
    maintenance.reindex()
    response.write = original


@implementer(IRegistry)
class CollectiveSolrMockRegistry(object):
    def __init__(self):
        self.active = False
        self.host = "localhost"
        self.port = None
        self.base = None
        self.async_indexing = False
        self.auto_commit = True
        self.commit_within = 0
        self.index_timeout = 0
        self.search_timeout = 0
        self.max_results = 0
        self.required = ()
        self.search_pattern = None
        self.facets = []
        self.filter_queries = ()
        self.slow_query_threshold = 0
        self.effective_steps = 1
        self.exclude_user = False
        self.field_list = []
        self.atomic_updates = False
        self.boost_script = ""
        self.solr_login = None
        self.solr_password = None
        self.stopwords = ""
        self.stopwords_case_insensitive = False

    def __getitem__(self, name):
        name_parts = name.split(".")
        return getattr(self, name_parts[2])

    def get(self, name, default=None):
        return

    def __setitem__(self, name, value):
        return

    def __contains__(self, name):
        return

    @property
    def records(self):
        return

    # Schema interface API

    def forInterface(self, interface, check=True, omit=(), prefix=None, factory=None):
        return self

    def registerInterface(self, interface, omit=(), prefix=None):
        return

    def collectionOfInterface(
        self, interface, check=True, omit=(), prefix=None, factory=None
    ):
        return


class CollectiveSolrMockRegistryLayer(Layer):
    """Solr test layer that fires up and shuts down a Solr instance. This
    layer can be used to unit test a Solr configuration without having to
    fire up Plone.
    """

    def setUp(self):
        provideUtility(
            provides=IRegistry, component=CollectiveSolrMockRegistry(), name=""
        )

    def tearDown(self):
        # XXX: we have to unregister the utility, otherwise the test fixture
        # will bleed into other tests. This currently makes a few unit tests
        # pass. We need to fix this properly before merging though. (timo)
        pass


def set_attributes(context, values):  # pragma: no cover
    for key, value in six.iteritems(values):
        setattr(context, key, value)


COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE = CollectiveSolrMockRegistryLayer()

COLLECTIVE_SOLR_FIXTURE = CollectiveSolrLayer(solr_active=True)

COLLECTIVE_SOLR_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_SOLR_FIXTURE,), name="CollectiveSolr:Integration"
)

COLLECTIVE_SOLR_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(COLLECTIVE_SOLR_FIXTURE,), name="CollectiveSolr:Functional"
)

COLLECTIVE_SOLR_ROBOT_TESTING = FunctionalTesting(
    bases=(COLLECTIVE_SOLR_FIXTURE, REMOTE_LIBRARY_BUNDLE_FIXTURE, z2.ZSERVER_FIXTURE),
    name="CollectiveSolr:Acceptance",
)

LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING = IntegrationTesting(
    bases=(LEGACY_COLLECTIVE_SOLR_FIXTURE,), name="CollectiveSolr:Integration"
)

LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(LEGACY_COLLECTIVE_SOLR_FIXTURE,), name="CollectiveSolr:Functional"
)



from unittest import TestCase

import six
from collective.solr.parser import SolrResponse, SolrSchema, parseDate
from collective.solr.tests.utils import getData
from DateTime import DateTime


class ParserTests(TestCase):
    def testParseSimpleSearchResults(self):
        search_response = getData("search_response.txt").decode("utf-8")
        response = SolrResponse(search_response.split("\n\n", 1)[1])
        results = response.response  # the result set is named 'response'
        self.assertEqual(results.numFound, "1")
        self.assertEqual(results.start, "0")
        self.assertEqual(response.actual_result_count, 1)
        match = results[0]
        self.assertEqual(len(results), 1)
        self.assertEqual(match.id, "500")
        self.assertEqual(match.name, "python test doc")
        self.assertEqual(match.popularity, 0)
        self.assertEqual(match.sku, "500")
        self.assertEqual(
            match.timestamp.ISO8601(), DateTime("2008-02-29 16:11:46.998 GMT").ISO8601()
        )
        headers = response.responseHeader
        self.assertEqual(headers["status"], 0)
        self.assertEqual(headers["QTime"], 0)
        self.assertEqual(headers["params"]["wt"], "xml")
        self.assertEqual(headers["params"]["indent"], "on")
        self.assertEqual(headers["params"]["rows"], "10")
        self.assertEqual(headers["params"]["q"], "id:[* TO *]")

    def testParseComplexSearchResults(self):
        complex_xml_response = getData("complex_xml_response.txt")
        response = SolrResponse(complex_xml_response)
        results = response.response  # the result set is named 'response'
        self.assertEqual(results.numFound, "2")
        self.assertEqual(response.actual_result_count, 2)
        self.assertEqual(results.start, "0")
        self.assertEqual(len(results), 2)
        first = results[0]
        self.assertEqual(first.cat, ["software", "search"])
        self.assertEqual(len(first.features), 7)
        self.assertEqual(
            [type(x).__name__ for x in first.features],
            ["str"] * 6 + [six.text_type.__name__],
        )
        self.assertEqual(first.id, "SOLR1000")
        self.assertEqual(first.inStock, True)
        self.assertEqual(
            first.incubationdate_dt.ISO8601(), DateTime("2006/01/17 GMT").ISO8601()
        )
        self.assertEqual(first.manu, "Apache Software Foundation")
        self.assertEqual(first.popularity, 10)
        self.assertEqual(first.price, 0.0)
        headers = response.responseHeader
        self.assertEqual(headers["status"], 0)
        self.assertEqual(headers["QTime"], 0)
        self.assertEqual(headers["params"]["indent"], "on")
        self.assertEqual(headers["params"]["rows"], "10")
        self.assertEqual(headers["params"]["start"], "0")
        self.assertEqual(headers["params"]["q"], "id:[* TO *]")
        self.assertEqual(headers["params"]["version"], "2.2")

    def testParseFacetSearchResults(self):
        facet_xml_response = getData("facet_xml_response.txt").decode("utf-8")
        response = SolrResponse(facet_xml_response)
        results = response.response  # the result set is named 'response'
        self.assertEqual(results.numFound, "1")
        self.assertEqual(results.start, "0")
        self.assertEqual(len(results), 0)
        headers = response.responseHeader
        self.assertEqual(type(headers), type({}))
        self.assertEqual(headers["status"], 0)
        self.assertEqual(headers["QTime"], 1)
        self.assertEqual(headers["params"]["facet.limit"], "-1")
        self.assertEqual(headers["params"]["rows"], "0")
        self.assertEqual(headers["params"]["facet"], "true")
        self.assertEqual(headers["params"]["facet.field"], ["cat", "inStock"])
        self.assertEqual(headers["params"]["indent"], "10")
        self.assertEqual(headers["params"]["q"], "solr")
        counts = response.facet_counts
        self.assertEqual(type(counts), type({}))
        self.assertEqual(counts["facet_queries"], {})
        self.assertEqual(counts["facet_fields"]["cat"]["electronics"], 0)
        self.assertEqual(counts["facet_fields"]["cat"]["monitor"], 0)
        self.assertEqual(counts["facet_fields"]["cat"]["search"], 1)
        self.assertEqual(counts["facet_fields"]["cat"]["software"], 1)
        self.assertEqual(counts["facet_fields"]["inStock"]["true"], 1)

    def testParseDateFacetSearchResults(self):
        facet_xml_response = getData("date_facet_xml_response.txt").decode("utf-8")
        response = SolrResponse(facet_xml_response)
        results = response.response  # the result set is named 'response'
        self.assertEqual(results.numFound, "42")
        self.assertEqual(results.start, "0")
        self.assertEqual(len(results), 0)
        headers = response.responseHeader
        self.assertEqual(type(headers), type({}))
        self.assertEqual(headers["status"], 0)
        self.assertEqual(headers["QTime"], 5)
        self.assertEqual(headers["params"]["facet.date"], "timestamp")
        self.assertEqual(headers["params"]["facet.date.start"], "NOW/DAY-5DAYS")
        self.assertEqual(headers["params"]["facet.date.end"], "NOW/DAY+1DAY")
        self.assertEqual(headers["params"]["facet.date.gap"], "+1DAY")
        self.assertEqual(headers["params"]["rows"], "0")
        self.assertEqual(headers["params"]["facet"], "true")
        self.assertEqual(headers["params"]["indent"], "true")
        self.assertEqual(headers["params"]["q"], "*:*")
        counts = response.facet_counts
        self.assertEqual(type(counts), type({}))
        self.assertEqual(counts["facet_queries"], {})
        self.assertEqual(counts["facet_fields"], {})
        timestamps = counts["facet_dates"]["timestamp"]
        self.assertEqual(timestamps["2007-08-11T00:00:00.000Z"], 1)
        self.assertEqual(timestamps["2007-08-12T00:00:00.000Z"], 5)
        self.assertEqual(timestamps["2007-08-13T00:00:00.000Z"], 3)
        self.assertEqual(timestamps["2007-08-14T00:00:00.000Z"], 7)
        self.assertEqual(timestamps["2007-08-15T00:00:00.000Z"], 2)
        self.assertEqual(timestamps["2007-08-16T00:00:00.000Z"], 16)
        self.assertEqual(timestamps["gap"], "+1DAY")
        self.assertEqual(
            timestamps["end"].ISO8601(), DateTime("2007-08-17 GMT").ISO8601()
        )

    def testParseConfig(self):
        schema_xml = getData("schema.xml").decode("utf-8")
        schema = SolrSchema(schema_xml.split("\n\n", 1)[1])
        self.assertEqual(len(schema), 22)  # 22 items defined in schema.xml
        self.assertEqual(schema["defaultSearchField"], "text")
        self.assertEqual(schema["uniqueKey"], "id")
        self.assertEqual(schema["solrQueryParser"].defaultOperator, "OR")
        self.assertEqual(schema["requiredFields"], ["id", "name"])
        self.assertEqual(schema["id"].type, "string")
        self.assertEqual(schema["id"].class_, "solr.StrField")
        self.assertEqual(schema["id"].required, True)
        self.assertEqual(schema["id"].omitNorms, True)
        self.assertEqual(schema["id"].multiValued, False)
        self.assertEqual(schema["cat"].class_, "solr.TextField")
        self.assertEqual(schema["cat"].required, False)
        self.assertEqual(schema["cat"].multiValued, True)
        self.assertEqual(schema["cat"].termVectors, True)
        self.assertEqual(schema["sku"].positionIncrementGap, "100")
        self.assertEqual(schema.features.multiValued, True)
        self.assertEqual(schema.weight.class_, "solr.SortableFloatField")
        self.assertEqual(schema.popularity.class_, "solr.SortableIntField")
        self.assertEqual(schema.inStock.class_, "solr.BoolField")
        self.assertEqual(schema.timestamp.class_, "solr.DateField")
        self.assertEqual(schema.timestamp.default, "NOW")
        self.assertEqual(schema.timestamp.multiValued, False)
        self.assertEqual(schema.timestamp.indexed, True)
        self.assertEqual(schema.word.indexed, False)
        fields = list(schema.values())
        self.assertEqual(len([f for f in fields if getattr(f, "required", False)]), 2)
        self.assertEqual(
            len([f for f in fields if getattr(f, "multiValued", False)]), 3
        )

    def testParseQuirkyResponse(self):
        quirky_response = getData("quirky_response.txt").decode("utf-8")
        response = SolrResponse(quirky_response)
        results = response.response  # the result set is named 'response'
        empty_uid = [r for r in results if r.UID == ""]
        self.assertEqual(empty_uid, [])

    def testParseResultsActualResultCount(self):
        complex_xml_response = getData("complex_xml_response.txt")
        response = SolrResponse(complex_xml_response)
        self.assertEqual(response.actual_result_count, 2)
        response.actual_result_count = 1
        results = response.response  # the result set is named 'response'
        self.assertEqual(response.actual_result_count, 1)
        self.assertEqual(len(results), 2)


class ParseDateHelperTests(TestCase):
    def testParseDateHelper(self):
        self.assertEqual(
            parseDate("2007-08-11T00:00:00.000Z").ISO8601(),
            DateTime(2007, 8, 11, 0, 0, 0, "GMT").ISO8601(),
        )
        self.assertEqual(
            parseDate("1900-01-01T00:00:00.000Z").ISO8601(),
            DateTime(1900, 1, 1, 0, 0, 0, "GMT").ISO8601(),
        )
        self.assertEqual(
            parseDate("999-12-31T00:00:00.000Z").ISO8601(),
            DateTime(999, 12, 31, 0, 0, 0, "GMT").ISO8601(),
        )
        self.assertEqual(
            parseDate("99-12-31T00:00:00.000Z").ISO8601(),
            DateTime("0099-12-31T00:00:00.000Z").ISO8601(),
        )



import six
from collective.solr import SolrMessageFactory
from collective.solr.interfaces import (
    IFacetTitleVocabularyFactory,
    ISolrConnectionManager,
)
from plone.registry.interfaces import IRegistry
from zope.component import getUtility, queryUtility
from zope.interface import implementer
from zope.schema.interfaces import IBaseVocabulary, IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary


@implementer(IVocabularyFactory)
class SolrIndexes(object):
    """vocabulary provider yielding all available solr indexes"""

    def __call__(self, context):
        items = []
        manager = queryUtility(ISolrConnectionManager)
        if manager is not None:
            schema = manager.getSchema()
            if schema is not None:
                for name, info in sorted(schema.items()):
                    if "indexed" in info and info.get("indexed", False):
                        items.append(name)
        if not items:
            registry = getUtility(IRegistry)
            items = registry["collective.solr.filter_queries"]
        return SimpleVocabulary([SimpleTerm(item) for item in items])


@implementer(IBaseVocabulary)
class I18NFacetTitles(object):
    """Vocabulary that wraps any term into message ids in the solr translation
    domain. This is generally the default behaviour for facet titles.
    """

    def __contains__(self, term):
        return True

    def getTerm(self, term):
        value = term
        title = SolrMessageFactory(term)
        if isinstance(term, six.text_type):
            # Terms must be byte strings
            term = term.encode("utf8")
        return SimpleTerm(value, term, title)


@implementer(IFacetTitleVocabularyFactory)
class I18NFacetTitlesVocabularyFactory(object):
    def __call__(self, context):
        return I18NFacetTitles()



from collective.solr.interfaces import ISearchDispatcher
from DateTime import DateTime
from Products.CMFCore.utils import _getAuthenticatedUser
from Products.CMFPlone.CatalogTool import CatalogTool
from zope.component import queryAdapter


def searchResults(self, REQUEST=None, **kw):
    """based on the version in `CMFPlone/CatalogTool.py`"""
    kw = kw.copy()
    only_active = not kw.get("show_inactive", False)
    user = _getAuthenticatedUser(self)
    kw["allowedRolesAndUsers"] = self._listAllowedRolesAndUsers(user)
    if only_active and not self.allow_inactive(kw):
        kw["effectiveRange"] = DateTime()

    adapter = queryAdapter(self, ISearchDispatcher)
    if adapter is not None:
        return adapter(REQUEST, **kw)
    else:
        return self._cs_old_searchResults(REQUEST, **kw)


def patchCatalogTool():
    """monkey patch plone's catalogtool with the solr dispatcher"""
    if not hasattr(CatalogTool, "_cs_old_searchResults"):
        CatalogTool._cs_old_searchResults = CatalogTool.searchResults
        CatalogTool.searchResults = searchResults
        CatalogTool.__call__ = searchResults



# -*- coding: utf-8 -*-

from collective.solr.dispatcher import solrSearchResults
from collective.solr.interfaces import ISearch
from collective.solr.interfaces import ISolrConnectionManager
from collective.solr.local import getLocal
from collective.solr.manager import SolrConnectionManager
from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING
from collective.solr.testing import activateAndReindex
from collective.solr.utils import activate
from zope.component import getSiteManager
from zope.component import getUtility
from zope.component import queryUtility

import unittest


class SolrMulticoreTests(unittest.TestCase):
    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def _close_connections(self, keys):
        for key in keys:
            key = key is not None and key or ""
            manager = getUtility(ISolrConnectionManager, name=key)
            manager.closeConnection(clearSchema=True)

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        activateAndReindex(self.portal)
        self._close_connections(keys=(None,))
        sm = getSiteManager()
        self._plone_manager = SolrConnectionManager()
        sm.registerUtility(self._plone_manager, ISolrConnectionManager, name="plone")
        self._foo_manager = SolrConnectionManager()
        sm.registerUtility(self._foo_manager, ISolrConnectionManager, name="foo")

    def tearDown(self):
        activate(active=False)
        self._close_connections((None, "plone", "foo"))
        sm = getSiteManager()
        sm.unregisterUtility(self._plone_manager, ISolrConnectionManager, name="plone")
        sm.unregisterUtility(self._foo_manager, ISolrConnectionManager, name="foo")

    def test_default_utility(self):
        manager = queryUtility(ISolrConnectionManager)
        # Default manager must not have a core parameter
        self.assertIsNone(manager.core)
        self.assertEqual("connection", manager.connection_key)
        self.assertEqual("schema", manager.schema_key)

    def test_named_utilities(self):
        default_manager = queryUtility(ISolrConnectionManager)
        default_conn = default_manager.getConnection()
        default_schema = default_manager.getSchema()

        plone_manager = queryUtility(ISolrConnectionManager, name="plone")
        plone_manager.setCore("plone")
        self.assertEqual("plone", plone_manager.core)
        self.assertEqual("connection_plone", plone_manager.connection_key)
        self.assertEqual("schema_plone", plone_manager.schema_key)

        plone_conn = plone_manager.getConnection()
        self.assertTrue(default_conn != plone_conn)
        self.assertEqual("/solr/plone", plone_conn.solrBase)

        plone_schema = plone_manager.getSchema()
        self.assertEqual(default_schema, plone_schema)

        foo_manager = queryUtility(ISolrConnectionManager, name="foo")
        foo_manager.setCore("foo")
        self.assertEqual("foo", foo_manager.core)
        self.assertEqual("connection_foo", foo_manager.connection_key)
        self.assertEqual("schema_foo", foo_manager.schema_key)

        foo_conn = foo_manager.getConnection()
        self.assertTrue(default_conn != foo_conn)
        self.assertEqual("/solr/foo", foo_conn.solrBase)

    def test_search_on_core(self):
        # Verify that connections does not exist before search
        self.assertIsNone(getLocal("connection"))
        self.assertIsNone(getLocal("connection_plone"))

        search = getUtility(ISearch, context=self.portal)
        results_core = search("+Title:*Welcome to Plone*", core="plone").results()
        self.assertEqual(1, len(results_core))
        self.assertIsNone(getLocal("connection"))
        self.assertIsNotNone(getLocal("connection_plone"))

        results_default = search("+Title:*Welcome to Plone*").results()
        self.assertEqual(1, len(results_default))

        # Verify connections and schemas
        default_conn = getLocal("connection")
        plone_conn = getLocal("connection_plone")
        self.assertTrue(default_conn != plone_conn)
        self.assertIsNotNone(default_conn)
        self.assertIsNotNone(plone_conn)

    def test_solr_search_result_with_core(self):
        # Verify that connections and schemas does not exist before search
        self.assertIsNone(getLocal("connection"))
        self.assertIsNone(getLocal("schema"))
        self.assertIsNone(getLocal("connection_plone"))
        self.assertIsNone(getLocal("schema_plone"))

        results = solrSearchResults(
            self.request, Title="Welcome to Plone", use_solr=True, core="plone"
        )
        self.assertEqual(1, len(results))

        # Verify that only concerned schemas and connections are now defined
        self.assertIsNone(getLocal("connection"))
        self.assertIsNone(getLocal("schema"))
        self.assertIsNotNone(getLocal("connection_plone"))
        self.assertIsNotNone(getLocal("schema_plone"))



from copy import deepcopy
from logging import getLogger

from Acquisition import aq_base
from collective.solr.exceptions import FallBackException
from collective.solr.interfaces import IFlare, ISearch, ISearchDispatcher
from collective.solr.monkey import patchCatalogTool
from collective.solr.parser import SolrResponse
from collective.solr.utils import isActive, padResults
from Missing import MV
from plone.registry.interfaces import IRegistry
from Products.ZCatalog.ZCatalog import ZCatalog
from zope.component import getUtility, queryMultiAdapter, queryUtility
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces.http import IHTTPRequest

patchCatalogTool()  # patch catalog tool to use the dispatcher...


logger = getLogger("collective.solr.dispatcher")


@implementer(ISearchDispatcher)
class SearchDispatcher(object):
    """adapter for potentially dispatching a given query to an
    alternative search backend (instead of the portal catalog)"""

    def __init__(self, context):
        self.context = context

    def __call__(self, request, **keywords):
        """decide on a search backend and perform the given query"""
        if isActive():
            try:
                return solrSearchResults(request, **keywords)
            except FallBackException:
                pass
        if getattr(aq_base(self.context), "_cs_old_searchResults", None):
            return self.context._cs_old_searchResults(request, **keywords)
        return ZCatalog.searchResults(self.context, request, **keywords)


def solrSearchResults(request=None, **keywords):
    """perform a query using solr after translating the passed in
    parameters with portal catalog semantics"""
    site = getSite()
    search = queryUtility(ISearch, context=site)
    if search is None:
        logger.warning("No search utility found in site %s", site)
        raise FallBackException

    registry = getUtility(IRegistry)
    config_required = registry["collective.solr.required"]

    if request is None:
        # try to get a request instance, so that flares can be adapted to
        # ploneflares and urls can be converted into absolute ones etc;
        # however, in this case any arguments from the request are ignored
        args = deepcopy(keywords)
        request = getattr(site, "REQUEST", None)
    elif IHTTPRequest.providedBy(request):
        args = deepcopy(request.form)
        args.update(keywords)  # keywords take precedence
    else:
        assert isinstance(request, dict), request
        args = deepcopy(request)
        args.update(keywords)  # keywords take precedence
        # if request is a dict, we need the real request in order to
        # be able to adapt to plone flares
        request = getattr(site, "REQUEST", args)

    if "path" in args and "navtree" in args["path"]:
        raise FallBackException  # we can't handle navtree queries yet

    use_solr = args.get("use_solr", False)  # A special key to force Solr
    if not use_solr and config_required:
        required = set(config_required).intersection(args)
        if required:
            for key in required:
                if not args[key]:
                    raise FallBackException
        else:
            raise FallBackException

    core = args.get("core", None)
    query, params = search.buildQueryAndParameters(**args)

    if query != {}:
        __traceback_info__ = (query, params, args)
        if core is not None:
            params["core"] = core
        response = search(query, **params)
    else:
        return SolrResponse()

    def wrap(flare):
        """wrap a flare object with a helper class"""
        adapter = queryMultiAdapter((flare, request), IFlare)
        return adapter is not None and adapter or flare

    schema = search.getManager(core=core).getSchema() or {}
    results = response.results()
    for idx, flare in enumerate(results):
        flare = wrap(flare)
        for missing in set(schema.stored).difference(flare):
            flare[missing] = MV
        results[idx] = wrap(flare)
    padResults(results, **params)  # pad the batch
    return response



from logging import getLogger
from time import time

import six
from Missing import MV
from Products.CMFPlone.utils import safe_unicode
from six.moves import map
from zope.component import queryUtility
from zope.interface import implementer

from collective.solr.exceptions import SolrInactiveException
from collective.solr.interfaces import (
    ISearch,
    ISolrConnectionManager,
    MAX_RESULTS_SUPPORTED_BY_SOLR,
)
from collective.solr.mangler import (
    cleanupQueryParameters,
    mangleQuery,
    optimizeQueryParameters,
    subtractQueryParameters,
)
from collective.solr.parser import SolrResponse
from collective.solr.queryparser import quote, quote_iterable_item
from collective.solr.utils import getConfig, isWildCard, prepare_wildcard, prepareData

try:
    from Products.LinguaPlone.catalog import languageFilter
except ImportError:

    def languageFilter(args):
        pass


logger = getLogger("collective.solr.search")


@implementer(ISearch)
class Search(object):
    """a search utility for solr"""

    def __init__(self):
        self.manager = None
        self.config = None

    def getManager(self, core=None):
        if self.manager is None or self.manager.core != core:
            name = core is not None and core or ""
            self.manager = queryUtility(ISolrConnectionManager, name=name)
            self.manager.setCore(core)
        return self.manager

    def getConfig(self):
        if self.config is None:
            self.config = getConfig()
        return self.config

    def search(
        self,
        query,
        wt="xml",
        sow="true",
        lowercase_operator="true",
        default_operator="AND",
        **parameters,
    ):
        """perform a search with the given querystring and parameters"""
        start = time()
        core = parameters.pop("core", None) or None
        config = self.getConfig()
        manager = self.getManager(core=core)
        manager.setSearchTimeout()
        connection = manager.getConnection()
        if connection is None:
            raise SolrInactiveException
        parameters["wt"] = wt
        parameters["sow"] = sow  # split on whitespace
        parameters["lowercaseOperators"] = lowercase_operator
        parameters["q.op"] = default_operator
        if "rows" not in parameters:
            parameters["rows"] = config.max_results or 10_000_000
            # Check if rows param is 0 for backwards compatibility. Before
            # Solr 4 'rows = 0' meant that there is no limitation. Solr 4
            # always expects a rows param > 0 though:
            # http://wiki.apache.org/solr/CommonQueryParameters#rows
            if parameters["rows"] == 0:
                parameters["rows"] = 10_000_000
            logger.debug(
                'falling back to "max_results" (%d) without a "rows" '
                "parameter: %r (%r)",
                config.max_results,
                query,
                parameters,
            )
        else:
            try:
                rows = int(parameters["rows"])
            except ValueError:
                # probably invalid, but we'll let solr complain about it
                pass
            else:
                if rows > MAX_RESULTS_SUPPORTED_BY_SOLR:
                    logger.warning(
                        'The "rows" parameter was set to "%s" but automatically limited '
                        " to %s to avoid a Solr java.lang.NumberFormatException",
                        parameters["rows"],
                        MAX_RESULTS_SUPPORTED_BY_SOLR,
                    )
                    parameters["rows"] = MAX_RESULTS_SUPPORTED_BY_SOLR
        if getattr(config, "highlight_fields", None):
            if parameters.get("hl", "false") == "true" and "hl.fl" not in parameters:
                parameters["hl"] = "true"
                parameters["hl.fl"] = config.highlight_fields or []
                parameters["hl.simple.pre"] = config.highlight_formatter_pre or " "
                parameters["hl.simple.post"] = config.highlight_formatter_post or " "
                parameters["hl.fragsize"] = (
                    getattr(config, "highlight_fragsize", None) or 100
                )
        if "fl" not in parameters:
            if config.field_list:
                parameters["fl"] = " ".join(config.field_list)
            else:
                parameters["fl"] = "* score"
        if isinstance(query, dict):
            query = " ".join([safe_unicode(val) for val in query.values()])
        logger.debug("searching for %r (%r)", query, parameters)
        if "sort" in parameters:  # issue warning for unknown sort indices
            index, order = parameters["sort"].split()
            schema = manager.getSchema() or {}
            field = schema.get(index, None)
            if field is None or not field.stored:
                logger.warning('sorting on non-stored attribute "%s"', index)
        response = connection.search(q=query, **parameters)
        results = SolrResponse(response)
        response.close()
        manager.setTimeout(None)
        elapsed = (time() - start) * 1000
        slow = config.slow_query_threshold
        if slow and elapsed >= slow:
            logger.info(
                "slow query: %d/%d ms for %r (%r)",
                results.responseHeader["QTime"],
                elapsed,
                query,
                parameters,
            )
        logger.debug("highlighting info: %s" % getattr(results, "highlighting", {}))
        return results

    __call__ = search

    def buildQueryAndParameters(self, default=None, **args):
        """helper to build a querystring for simple use-cases"""
        core = args.pop("core", None) or None
        schema = self.getManager(core=core).getSchema() or {}

        params = subtractQueryParameters(args)
        params = cleanupQueryParameters(params, schema)
        config = self.getConfig()

        languageFilter(args)
        prepareData(args)
        mangleQuery(args, config, schema)

        logger.debug('building query for "%r", %r', default, args)
        schema = self.getManager(core=core).getSchema() or {}
        # no default search field in Solr 7
        defaultSearchField = getattr(schema, "defaultSearchField", "SearchableText")
        args[None] = default
        query = {}

        for name, value in args.items():
            field_name = name
            if name and name[0] in "+-":
                field_name = name[1:]
            field = schema.get(field_name or defaultSearchField, None)
            if field is None or not field.indexed:
                logger.info(
                    'dropping unknown search attribute "%s" ' " (%r) for query: %r",
                    name,
                    value,
                    args,
                )
                continue
            if isinstance(value, bool):
                value = str(value).lower()
            elif not value:  # solr doesn't like empty fields (+foo:"")
                if not name:
                    continue
                logger.info(
                    'empty search term form "%s:%s", aborting buildQuery'
                    % (name, value)
                )
                return {}, params
            elif field.class_ == "solr.BoolField":
                if not isinstance(value, (tuple, list)):
                    value = [value]
                falses = "0", "False", MV
                true = lambda v: bool(v) and v not in falses
                value = set(map(true, value))
                if not len(value) == 1:
                    assert len(value) == 2  # just to make sure
                    continue  # skip when "true or false"
                value = str(value.pop()).lower()
            elif isinstance(value, (tuple, list)):
                # list items should be treated as literals, but
                # nevertheless only get quoted when necessary
                value = "(%s)" % " OR ".join(map(quote_iterable_item, value))
            elif isinstance(value, set):  # sets are taken literally
                if len(value) == 1:
                    query[name] = "".join(value)
                else:
                    query[name] = "(%s)" % " OR ".join(value)
                continue
            elif isinstance(value, six.string_types):
                if field.class_ == "solr.TextField":
                    if isWildCard(value):
                        value = prepare_wildcard(value)
                    value = quote(value, textfield=True)
                    # if we have an intra-word hyphen, we need quotes
                    if "\\-" in value or "\\+" in value:
                        if value[0] != '"':
                            value = '"%s"' % value
                else:
                    value = quote(value)
                if not value:  # don't search for empty strings, even quoted
                    continue
            elif isinstance(value, (int, float)):
                pass  # Do not raise an error for int or float values
            else:
                logger.info('skipping unsupported value "%r" (%s)', value, name)
                continue
            if name is None:
                if value and value[0] not in "+-":
                    value = "+%s" % value
            else:
                if name[0] not in "+-":
                    value = "+%s:%s" % (name, value)
                else:
                    value = "%s:%s" % (name, value)
            query[name] = value
        logger.debug('built query "%s"', query)

        if query:
            optimizeQueryParameters(query, params)
        return query, params



from Acquisition import aq_base
from plone import api
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.layout.icons.interfaces import IContentIcon
from plone.i18n.normalizer.interfaces import IIDNormalizer
from plone.registry.interfaces import IRegistry
from plone.uuid.interfaces import IUUID
from Products.CMFPlone.browser.ploneview import Plone as PloneView
from zope.component import getMultiAdapter, getUtility
from zope.globalrequest import getRequest
from zope.interface import implementer

missing = object()


@implementer(IContentListingObject)
class FlareContentListingObject(object):
    def __init__(self, flare):
        self.flare = flare

    def getId(self):
        return self.flare.getId

    def getObject(self):
        return self.flare.getObject()

    def getDataOrigin(self):
        return self.flare.getObject()

    def getPath(self):
        return self.flare.getPath()

    def getURL(self, relative=False):
        return self.flare.getURL(relative)

    def uuid(self):
        if "UID" in self.flare:
            return self.flare.UID
        else:
            return IUUID(self.getObject())

    def getIcon(self):
        return getMultiAdapter(
            (self.getObject(), getRequest(), self.flare), interface=IContentIcon
        )()

    def getSize(self):
        return self.flare.getObjSize

    def review_state(self):
        return self.flare.review_state

    def listCreators(self):
        return self.flare.listCreators

    def Creator(self):
        return self.flare.Creator

    def Subject(self):
        return self.flare.Subject

    def Publisher(self):
        raise NotImplementedError

    def listContributors(self):
        raise NotImplementedError

    def Contributors(self):
        return self.listContributors()

    def Date(self, zone=None):
        return self.flare.Date

    def CreationDate(self, zone=None):
        return self.flare.created

    def EffectiveDate(self, zone=None):
        # Work around an incompatibility of Archetypes/DateTime
        # in effective. See #13362
        return self.getObject().EffectiveDate()

    def ExpirationDate(self, zone=None):
        return self.flare.expires

    def ModificationDate(self, zone=None):
        return self.flare.modified

    def Format(self):
        raise NotImplementedError

    def Identifier(self):
        return self.getURL()

    def Language(self):
        return self.flare.Language

    def Rights(self):
        raise NotImplementedError

    def Title(self):
        return self.flare.Title

    def Description(self):
        return self.flare.Description

    def Type(self):
        return self.flare.Type

    def ContentTypeClass(self):
        return "contenttype-" + getUtility(IIDNormalizer).normalize(self.PortalType())

    def PortalType(self):
        return self.flare.portal_type

    def Author(self):
        return self.getUserData(self.Creator())

    def getUserData(self, username):
        request = getRequest()
        _usercache = request.get("usercache", None)
        if _usercache is None:
            request.set("usercache", {})
            _usercache = {}
        userdata = _usercache.get(username, None)
        if userdata is None:
            membershiptool = api.portal.get_tool("portal_membership")
            userdata = membershiptool.getMemberInfo(self.Creator())
            if not userdata:
                userdata = {
                    "username": username,
                    "description": "",
                    "language": "",
                    # TODO
                    # string:${navigation_root_url}/author/${item_creator}
                    "home_page": "/HOMEPAGEURL",
                    "location": "",
                    "fullname": username,
                }
            request.usercache[username] = userdata
        return userdata

    def CroppedDescription(self):
        registry = getUtility(IRegistry)
        length = registry.get("plone.search_results_description_length")
        plone_view = PloneView(None, None)
        if not length or not isinstance(length, int):
            # fallback if registry key is None
            length = 160
        return plone_view.cropText(self.flare.Description, length)

    @property
    def UID(self):
        # Alias for when this is used like a brain
        return self.flare.UID

    @property
    def modified(self):
        # Alias for when this is used like a brain
        return self.ModificationDate()

    @property
    def portal_type(self):
        # Alias for when this is used like a brain
        return self.PortalType()

    def MimeTypeIcon(self):
        raise NotImplementedError

    def image_scales(self):
        return self.flare.image_scales

    def __getattr__(self, name):
        """We'll override getattr so that we can defer name lookups to the real
        underlying objects without knowing the names of all attributes.

        Adapted from plone.app.contentlisting.catalog.CatalogContentListingObject
        """
        if name.startswith("_"):
            raise AttributeError(name)
        flare_value = self.flare.get(name, missing)
        if flare_value is not missing:
            return flare_value
        object_value = getattr(aq_base(self.getObject()), name, missing)
        if object_value is not missing:
            return object_value
        raise AttributeError(name)
